# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# Copyright (C) 2017, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import time
import subprocess
from enum import Enum
import ldap
from lib389._mapped_object import DSLdapObject, DSLdapObjects, _gen_or, _gen_filter, _term_gen
from lib389._constants import SER_ROOT_DN, SER_ROOT_PW
from lib389.utils import gentime_to_posix_time, gentime_to_datetime
from lib389.plugins import AccountPolicyPlugin, AccountPolicyConfig, AccountPolicyEntry
from lib389.cos import CosTemplates
from lib389.mappingTree import MappingTrees
from lib389.idm.role import Roles
from lib389.extended_operations import LdapSSOTokenRequest, LdapSSOTokenResponse


class AccountState(Enum):
    ACTIVATED = "activated"
    DIRECTLY_LOCKED = "directly locked through nsAccountLock"
    # TODO: Indirectly locked - revise the UI check
    INDIRECTLY_LOCKED = "indirectly locked through a Role"
    INACTIVITY_LIMIT_EXCEEDED = "inactivity limit exceeded"

    def describe(self, role_dn=None):
        if self.name == "INDIRECTLY_LOCKED" and role_dn is not None:
            return f'{self.value} - {role_dn}'
        else:
            return f'{self.value}'


class Account(DSLdapObject):
    """A single instance of Account entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(Account, self).__init__(instance, dn)
        self._protected = False

    def _format_status_message(self, message, create_time, modify_time, last_login_time, limit, role_dn=None):
        params = {}
        now = time.mktime(time.gmtime())
        if create_time:
            params["Creation Date"] = gentime_to_datetime(create_time)
        if modify_time:
            params["Modification Date"] = gentime_to_datetime(modify_time)
        params["Last Login Date"] = None
        params["Time Until Inactive"] = None
        params["Time Since Inactive"] = None
        if last_login_time:
            params["Last Login Date"] = gentime_to_datetime(last_login_time)
            if limit:
                remaining_time = float(limit) + gentime_to_posix_time(last_login_time) - now
                if remaining_time <= 0:
                    if message == AccountState.INACTIVITY_LIMIT_EXCEEDED:
                        params["Time Since Inactive"] = remaining_time
                else:
                    params["Time Until Inactive"] = remaining_time
        result = {"state": message, "params": params, "calc_time": now, "role_dn": None}
        if role_dn is not None:
            result["role_dn"] = role_dn
        return result

    def _dict_get_with_ignore_indexerror(self, dict, attr):
        try:
            return dict[attr][0]
        except IndexError:
            return ""

    def status(self):
        """Check if account is locked by Account Policy plugin or
        nsAccountLock (directly or indirectly)

        :returns: a dict in a format -
                  {"status": status, "params": activity_data, "calc_time": epoch_time}
        """

        inst = self._instance

        # Fetch Account Policy data if its enabled
        plugin = AccountPolicyPlugin(inst)
        try:
            config_dn = plugin.get_attr_val_utf8("nsslapd-pluginarg0")
        except IndexError:
            self._log.debug("The bound user doesn't have rights to access Account Policy settings. Not checking.")
        state_attr = ""
        alt_state_attr = ""
        limit = ""
        spec_attr = ""
        limit_attr = ""
        process_account_policy = False
        mapping_trees = MappingTrees(inst)
        try:
            root_suffix = mapping_trees.get_root_suffix_by_entry(self.dn)
            if str.lower(root_suffix) == str.lower(self.dn):
                raise ValueError("Root suffix can't be locked or unlocked via dsidm functionality.")
        except ldap.NO_SUCH_OBJECT:
            self._log.debug("Can't acquire root suffix from user DN. Probably - insufficient rights. Skipping this step.")
        try:
            process_account_policy = plugin.status()
        except IndexError:
            pass

        if process_account_policy and config_dn is not None:
            config = AccountPolicyConfig(inst, config_dn)
            config_settings = config.get_attrs_vals_utf8(["stateattrname", "altstateattrname",
                                                          "specattrname", "limitattrname"])
            state_attr = self._dict_get_with_ignore_indexerror(config_settings, "stateattrname")
            alt_state_attr = self._dict_get_with_ignore_indexerror(config_settings, "altstateattrname")
            spec_attr = self._dict_get_with_ignore_indexerror(config_settings, "specattrname")
            limit_attr = self._dict_get_with_ignore_indexerror(config_settings, "limitattrname")

            mapping_trees = MappingTrees(inst)
            root_suffix = mapping_trees.get_root_suffix_by_entry(self.dn)
            cos_entries = CosTemplates(inst, root_suffix)
            accpol_entry_dn = ""
            for cos in cos_entries.list():
                if cos.present(spec_attr):
                    accpol_entry_dn = cos.get_attr_val_utf8_l(spec_attr)
            if accpol_entry_dn:
                accpol_entry = AccountPolicyEntry(inst, accpol_entry_dn)
            else:
                accpol_entry = config
            limit = accpol_entry.get_attr_val_utf8_l(limit_attr)

        # Fetch account data
        account_data = self.get_attrs_vals_utf8(["createTimestamp", "modifyTimeStamp",
                                                 "nsAccountLock", state_attr])

        last_login_time = self._dict_get_with_ignore_indexerror(account_data, state_attr)
        # if last_login_time not exist then check alt_state_attr only if its not disabled and exist
        if not last_login_time and alt_state_attr in account_data:
            last_login_time = self._dict_get_with_ignore_indexerror(account_data, alt_state_attr)

        create_time = self._dict_get_with_ignore_indexerror(account_data, "createTimestamp")
        modify_time = self._dict_get_with_ignore_indexerror(account_data, "modifyTimeStamp")

        acct_roles = self.get_attr_vals_utf8_l("nsRole")
        mapping_trees = MappingTrees(inst)
        root_suffix = ""
        try:
            root_suffix = mapping_trees.get_root_suffix_by_entry(self.dn)
        except ldap.NO_SUCH_OBJECT:
            self._log.debug("The bound user doesn't have rights to access disabled roles settings. Not checking.")
        if root_suffix:
            roles = Roles(inst, root_suffix)
            try:
                disabled_roles = roles.get_disabled_roles()

                # Locked indirectly through a role
                locked_indirectly_role_dn = ""
                for role in acct_roles:
                    if str.lower(role) in [str.lower(role.dn) for role in disabled_roles.keys()]:
                        locked_indirectly_role_dn = role
                if locked_indirectly_role_dn:
                    return self._format_status_message(AccountState.INDIRECTLY_LOCKED, create_time, modify_time,
                                                       last_login_time, limit, locked_indirectly_role_dn)
            except ldap.NO_SUCH_OBJECT:
                pass

        # Locked directly
        if self._dict_get_with_ignore_indexerror(account_data, "nsAccountLock") == "true":
            return self._format_status_message(AccountState.DIRECTLY_LOCKED,
                                               create_time, modify_time, last_login_time, limit)

        # Locked indirectly through Account Policy plugin
        if process_account_policy and last_login_time:
            # Now check the Account Policy Plugin inactivity limits
            remaining_time = float(limit) - (time.mktime(time.gmtime()) - gentime_to_posix_time(last_login_time))
            if remaining_time <= 0:
                return self._format_status_message(AccountState.INACTIVITY_LIMIT_EXCEEDED,
                                                   create_time, modify_time, last_login_time, limit)
        # All checks are passed - we are active
        return self._format_status_message(AccountState.ACTIVATED, create_time, modify_time, last_login_time, limit)

    def ensure_lock(self):
        """Ensure nsAccountLock is set to 'true'"""

        self.replace('nsAccountLock', 'true')

    def ensure_unlock(self):
        """Unset nsAccountLock if it's set"""

        self.ensure_removed('nsAccountLock', None)

    def lock(self):
        """Set nsAccountLock to 'true'"""

        current_status = self.status()
        if current_status["state"] == AccountState.DIRECTLY_LOCKED:
            raise ValueError("Account is already locked")
        self.replace('nsAccountLock', 'true')

    def unlock(self):
        """Unset nsAccountLock if it's set and reset lastLoginTime if account is locked due to inactivity"""

        current_status = self.status()

        if current_status["state"] == AccountState.ACTIVATED:
            raise ValueError("Account is already active")

        if current_status["state"] == AccountState.DIRECTLY_LOCKED:
            # Account is directly locked with nsAccountLock attribute
            self.remove('nsAccountLock', None)
        elif current_status["state"] == AccountState.INACTIVITY_LIMIT_EXCEEDED:
            # Account is locked due to inactivity - reset lastLoginTime to current time
            # The lastLoginTime attribute stores its value in GMT/UTC time (Zulu time zone)
            current_time = time.strftime('%Y%m%d%H%M%SZ', time.gmtime())
            self.replace('lastLoginTime', current_time)
        elif current_status["state"] == AccountState.INDIRECTLY_LOCKED:
            # Account is locked through a role
            role_dn = current_status.get("role_dn")
            if role_dn:
                raise ValueError(f"Account is locked through role {role_dn}. "
                                 f"Please modify the role to unlock this account.")
            else:
                raise ValueError("Account is locked through an unknown role. "
                                 "Please check the roles configuration to unlock this account.")
        else:
            # Should not happen, but just in case
            raise ValueError(f"Unknown lock state: {current_status['state'].value}")

    # If the account can be bound to, this will attempt to do so. We don't check
    # for exceptions, just pass them back!
    def bind(self, password=None, *args, **kwargs):
        """Open a new connection and bind with the entry.
        You can pass arguments that will be passed to openConnection.

        :param password: An entry password
        :type password: str
        :returns: Connection with a binding as the entry
        """

        inst_clone = self._instance.clone({SER_ROOT_DN: self.dn, SER_ROOT_PW: password})
        inst_clone.open(*args, **kwargs)
        return inst_clone

    def rebind(self, password):
        """Rebind on the same connection
        :param password: An entry password
        :type password: str
        """
        self._instance.simple_bind_s(self.dn, password, escapehatch='i am sure')

    def sasl_bind(self, *args, **kwargs):
        """Open a new connection and bind with the entry via SASL.
        You can pass arguments that will be pass to clone.

        :return: Connection with a sasl binding to the entry.
        """
        inst_clone = self._instance.clone({SER_ROOT_DN: self.dn})
        inst_clone.open(*args, **kwargs)
        return inst_clone

    def create_keytab(self):
        """
        Create a keytab for this account valid to bind with.
        """
        assert self._instance.realm is not None

        myuid = self.get_attr_val_utf8('uid')
        self._instance.realm.create_principal(myuid)
        self._instance.realm.create_keytab(myuid, "/tmp/%s.keytab" % myuid)

        self._keytab = "/tmp/%s.keytab" % myuid

    def bind_gssapi(self):
        """
        Bind this account with gssapi credntials (if available)
        """
        assert self._instance.realm is not None
        # Kill any local kerberos ccache.
        subprocess.call(['kdestroy', '-A'])

        # This uses an in memory once off ccache.
        os.environ["KRB5_CLIENT_KTNAME"] = self._keytab

        # Because of the way that GSSAPI works, we can't
        # use the normal dirsrv open method.
        inst_clone = self._instance.clone()
        inst_clone.open(saslmethod='gssapi')
        return inst_clone

    def enroll_certificate(self, der_path):
        """Enroll a certificate for certmap verification. Because of the userCertificate
        attribute, we have to do this on userAccount which has support for it.

        :param der_path: the certificate file in DER format to include.
        :type der_path: str
        """
        # Given a cert path, add this to the object as a userCertificate
        crt = None
        with open(der_path, 'rb') as f:
            crt = f.read()
        self.add('usercertificate;binary', crt)

    def reset_password(self, new_password):
        """Set the password of the account: This requires write permission to
        the userPassword attribute, so likely is only possible as an administrator
        of the directory.

        :param new_password: The new password value to set
        :type new_password: str
        """
        self.set('userPassword', new_password)

    def change_password(self, current_password, new_password):
        """Using the accounts current bind password, performan an ldap passwd
        change extended operation. This does not required elevated permissions
        to read/write the userPassword field, so is the way that most accounts
        would change their password. This doesn't work on all classes of objects
        so it could error.

        :param current_password: The existing password value
        :type current_password: str
        :param new_password: The new password value to set
        :type new_password: str
        """
        # Please see _mapped_object.py and DSLdapObject for why this is structured
        # in this way re-controls.
        self._instance.passwd_s(self._dn, current_password, new_password,
            serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')

    def request_sso_token(self):
        """From an authenticated connection, request a single sign on token (cookie) for
        future use.

        Note this function will fail if the connection was itself authenticated with a
        token - this is to prevent token renewal from being infinite which may be a
        security risk.

        :returns: String of the token for use with authenticate_sso_token
        """
        # Build the request
        req = LdapSSOTokenRequest()
        # Get the response
        (_, res) = self._instance.extop_s(req, serverctrls=self._server_controls,
            clientctrls=self._client_controls, escapehatch='i am sure')
        (lt, token) = LdapSSOTokenResponse(res).responseValue
        return token

    def authenticate_sso_token(self, token, *args, **kwargs):
        """
        Given an entry (with dn), bind using this authentication token.

        :param token: An entry token
        :type token: str
        :returns: Connection with a binding as the entry
        """
        return self.bind(token, *args, **kwargs)


class Accounts(DSLdapObjects):
    """DSLdapObjects that represents Account entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all account entries below
    :type basedn: str
    """

    def __init__(self, instance, basedn):
        super(Accounts, self).__init__(instance)
        # These are all the objects capable of holding a password.
        self._objectclasses = [
            'nsAccount',
            'nsPerson',
            'simpleSecurityObject',
            'organization',
            'person',
            'account',
            'organizationalUnit',
            'netscapeServer',
            'domain',
            'posixAccount',
            'shadowAccount',
            'posixGroup',
            'mailRecipient',
        ]
        # MUST BE NONE. For more, see _gen_filter in _mapped_object.py.
        self._filterattrs = None
        self._childobject = Account
        self._basedn = basedn

    #### This is copied from DSLdapObjects, but change _gen_and to _gen_or!!!

    def _get_objectclass_filter(self):
        return _gen_or(
            _gen_filter(_term_gen('objectclass'), self._objectclasses)
        )


class Anonymous(DSLdapObject):
    """A single instance of Anonymous bind

    :param instance: An instance
    :type instance: lib389.DirSrv
    """
    def __init__(self, instance):
        super(Anonymous, self).__init__(instance, dn=None)

    def bind(self, *args, **kwargs):
        """Open a new connection and Anonymous bind .
        You can pass arguments that will be passed to openConnection.
        :returns: Connection with a binding as the Anonymous
        """
        inst_clone = self._instance.clone({SER_ROOT_DN: '', SER_ROOT_PW: ''})
        inst_clone.open(*args, **kwargs)
        return inst_clone



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019, William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.idm.user import nsUserAccounts
from lib389.idm.group import Groups
from lib389.plugins import MemberOfPlugin
from lib389.utils import basedn_to_ldap_dns_uri
from lib389.cli_base import CustomHelpFormatter

SSSD_CONF_TEMPLATE = """
#
# sssd.conf
# Generated by 389 Directory Server - dsidm
#
# For more details see man sssd.conf and man sssd-ldap
# Be sure to review the content of this file to ensure it is secure and correct
# in your environment.
{ldap_uri_warning}

[domain/ldap]
# Uncomment this for more verbose logging.
# debug_level=3

# Cache hashes of user authentication for offline auth.
cache_credentials = True
id_provider = ldap
auth_provider = ldap
access_provider = ldap
chpass_provider = ldap
ldap_schema = {schema_type}
ldap_search_base = {basedn}
ldap_uri = {ldap_uri}
# If you have DNS SRV records, you can use the following instead. This derives
# from your ldap_search_base.
# ldap_uri = _srv_

ldap_tls_reqcert = demand
# To use cacert dir, place *.crt files in this path then run:
# /usr/bin/openssl rehash /etc/openldap/certs
# or (for older versions of openssl)
# /usr/bin/c_rehash /etc/openldap/certs
ldap_tls_cacertdir = /etc/openldap/certs

# Path to the cacert
# ldap_tls_cacert = /etc/openldap/certs/ca.crt

# Only users who match this filter can login and authorise to this machine. Note
# that users who do NOT match, will still have their uid/gid resolve, but they
# can't login.
{ldap_access_filter}

enumerate = false
access_provider = ldap
ldap_user_member_of = memberof
ldap_user_gecos = cn
ldap_user_uuid = nsUniqueId
ldap_group_uuid = nsUniqueId
# This is really important as it allows SSSD to respect nsAccountLock
ldap_account_expire_policy = rhds
ldap_access_order = filter, expire
# Setup for ssh keys
# Inside /etc/ssh/sshd_config add the lines:
#   AuthorizedKeysCommand /usr/bin/sss_ssh_authorizedkeys
#   AuthorizedKeysCommandUser nobody
# You can test with the command: sss_ssh_authorizedkeys <username>
ldap_user_ssh_public_key = nsSshPublicKey

# This prevents an issue where the Directory is recursively walked on group
# and user look ups. It makes the client faster and more responsive in almost
# every scenario.
ignore_group_members = False

[sssd]
services = nss, pam, ssh, sudo
config_file_version = 2

domains = ldap
[nss]
homedir_substring = /home

"""

def sssd_conf(inst, basedn, log, args):

    schema_type = "rfc2307"
    try:
        mo_plugin = MemberOfPlugin(inst)
        if mo_plugin.status():
            schema_type = "rfc2307bis"
    except:
        schema_type = "unknown - likely access denied to memberof plugin config"

    ldap_access_filter = "# ldap_access_filter = (memberOf=<dn>)"
    if args.allowed_group:
        groups = Groups(inst, basedn)
        g_access = groups.get(args.allowed_group)
        ldap_access_filter = 'ldap_access_filter = (memberOf=%s)' % g_access.dn

    ldap_uri_warning = ""
    if inst.ldapuri.startswith("ldapi://"):
        ldap_uri_warning = "# ⚠️  WARNING: ldap_uri starts with ldapi:// - you should review this parameter"

    # Print a customised sssd.config.
    print(SSSD_CONF_TEMPLATE.format(
        basedn=basedn,
        schema_type=schema_type,
        ldap_uri=inst.ldapuri,
        ldap_access_filter=ldap_access_filter,
        ldap_uri_warning=ldap_uri_warning
    ))

    # Print a customised sssd.conf to log for test purpose
    log.debug(SSSD_CONF_TEMPLATE.format(
        basedn=basedn,
        schema_type=schema_type,
        ldap_uri=inst.ldapuri,
        ldap_access_filter=ldap_access_filter,
        ldap_uri_warning=ldap_uri_warning
    ))

LDAP_CONF_TEMPLATE = """
#
# OpenLDAP client configuration
# Generated by 389 Directory Server - dsidm
#

# See ldap.conf(5) for details
# This file should be world readable but not world writable.

BASE    {basedn}
# Remember to check this: you can have multiple uris on this line. You may have
# multiple servers or load balancers in your environment.
URI     {ldap_uri}
# If you have DNS SRV records you can use:
# URI   {ldap_dns_uri}

DEREF   never
# To use cacert dir, place *.crt files in this path then run:
# /usr/bin/openssl rehash /etc/openldap/certs
# or (for older versions of openssl)
# /usr/bin/c_rehash /etc/openldap/certs
TLS_CACERTDIR /etc/openldap/certs
# TLS_CACERT /etc/openldap/certs/ca.crt

"""

def ldap_conf(inst, basedn, log, args):
    # Print a customised ldap.conf or ldaprc
    print(LDAP_CONF_TEMPLATE.format(
        basedn=basedn,
        ldap_uri=inst.ldapuri,
        ldap_dns_uri=basedn_to_ldap_dns_uri(basedn),
    ))

    # Print a customised ldap.conf to log for test purpose
    log.debug(LDAP_CONF_TEMPLATE.format(
        basedn=basedn,
        ldap_uri=inst.ldapuri,
        ldap_dns_uri=basedn_to_ldap_dns_uri(basedn),
    ))

DISPLAY_TEMPLATE = """
# This is a generic list of LDAP client configuration parameters you may require
# for connecting a client to this server. Some of them may or may not apply
# to your application, so consult your application documentation for further
# assistance.
#
# This program makes a number of assumptions about your data and configuration
# which may not be correct. Be sure to check these values for your situation.

; ldap uri
; This is the uri of the server you will connect to and authenticate to. It
; must be a valid subjectAltName in the presented TLS certificate. Note that this
; is not an exhaustive list of your LDAP servers, and other applications in your
; network like load balancers may affect this. This is just what we derive from
; your current connection.
ldap_uri = {ldap_uri}

; ldap dns discovery uri
; In some environments, you may have DNS SRV records such as
; "_ldap._tcp.<domain name>". If these are present in your dns server, you can
; use the following uri.
ldap_uri = {ldap_dns_uri}

; ca_cert
; To correctly use TLS, you require the valid CA cert that issued your LDAP TLS
; certificates. Sometimes a copy of this may be in your server instance as
ca_cert = /etc/dirsrv/slapd-<instance>/ca.crt
; However that's not guaranteed. You can show the certs from the LDAP server
; by sshing to the server and running:
certutil -L -d /etc/dirsrv/slapd-<instance>/
; If you can identify the CA certificate name, you can then view it with:
certutil -L -n <ca cert name> -a -d /etc/dirsrv/slapd-<instance>/
; This should be a pem file you can use in your application's CA.
; Some applications don't require a ca certificate parameter, and will use the
; ca certificate from /etc/openldap/ldap.conf. You should configure ldap.conf
; in these cases. See the 'client_config ldap.conf' command in dsidm.

; basedn
; The basedn is the root suffix where all searches will originate from for
; LDAP objects.
basedn = {basedn}

; schema_type
; LDAP servers have different ways to structure their objects and group
; relationships. Legacy servers will use rfc2307, where as modern servers will
; use rfc2307bis (requires MemberOf plugin to be enabled). This is the schema
; setting of your directory based on your running configuration (if we can
; detect it).
schema_type = {schema_type}

; user/account basedn
; Some applications may optionally use a user/account basedn to limit searches
; in the directory. This can be for performance or security reasons. Generally
; you shouldn't need this, preferring to use groups and filters for access
; control.
user_basedn = {user_basedn}

; user filter
; This is an ldap filter that will return only user objects. Additionally some
; applications will template into the filter (similar to sql statements) or they
; will generate the filter based on attributes. We list a number of possible
; filters you might use, but you should customise this for your application.
;
; If you are using rfc2307bis, you can use this filter to provide authorisation
; support by adding filters such as: (memberOf=<groupdn>)
user_filter = {user_filter}
user_filter = {user_id_filter}

; group basedn
; Some applications may optionnaly use a group basedn to limit searches in the
; directory. This can be for performance or security reasons. Generally you
; shouldn't need this, preferring to use groups and filters for access control.
group_basedn = {group_basedn}

; group filter
; This is an ldap filter that will return only group objects. Additionally
; some applications will template into the filter (similar to sql statements)
; or they will generate the filter base on attributes. We list a number of
; possible filters you might use, but you should customise this for your
; application.
group_filter = {group_filter}
group_filter = {group_id_filter}

; attribute mappings
; Due to the variety of schemas and attribute mappings in LDAP, there are
; different representations of attributes and values. This is a guess at
; the mappings that exist in your server, and what attributes you should
; configure and use.
unique id = {uuid_attr}
user rdn = {user_rdn}
user identifier = {user_rdn}
group rdn = {group_rdn}
group member attribute = {group_member}

"""


def display(inst, basedn, log, args):

    users = nsUserAccounts(inst, basedn)
    groups = Groups(inst, basedn)

    schema_type = "rfc2307"
    try:
        mo_plugin = MemberOfPlugin(inst)
        if mo_plugin.status():
            schema_type = "rfc2307bis"
    except:
        schema_type = "unknown - likely access denied to memberof plugin config"

    # Get required information
    print(DISPLAY_TEMPLATE.format(
        ldap_uri=inst.ldapuri,
        ldap_dns_uri=basedn_to_ldap_dns_uri(basedn),
        basedn=basedn,
        schema_type=schema_type,
        user_basedn=users._basedn,
        user_filter=users._get_objectclass_filter(),
        user_id_filter=users._get_selector_filter('<PARAM>'),
        group_basedn=groups._basedn,
        group_filter=groups._get_objectclass_filter(),
        group_id_filter=groups._get_selector_filter('<PARAM>'),
        uuid_attr='nsUniqueId',
        user_rdn=users._filterattrs[0],
        group_rdn=groups._filterattrs[0],
        group_member='member',
    ))

    # Print required information to log for test purpose
    log.debug(DISPLAY_TEMPLATE.format(
        ldap_uri=inst.ldapuri,
        ldap_dns_uri=basedn_to_ldap_dns_uri(basedn),
        basedn=basedn,
        schema_type=schema_type,
        user_basedn=users._basedn,
        user_filter=users._get_objectclass_filter(),
        user_id_filter=users._get_selector_filter('<PARAM>'),
        group_basedn=groups._basedn,
        group_filter=groups._get_objectclass_filter(),
        group_id_filter=groups._get_selector_filter('<PARAM>'),
        uuid_attr='nsUniqueId',
        user_rdn=users._filterattrs[0],
        group_rdn=groups._filterattrs[0],
        group_member='member',
    ))

def create_parser(subparsers):
    client_config_parser = subparsers.add_parser('client_config',
        help="Display and generate client example configs for this LDAP server")

    subcommands = client_config_parser.add_subparsers(help='action')

    sssd_conf_parser = subcommands.add_parser('sssd.conf',
        help="Generate a SSSD configuration for this LDAP server")
    # Allowed login groups for filter?
    sssd_conf_parser.add_argument('allowed_group', nargs='?', help="The name of the group allowed access to this system")
    sssd_conf_parser.set_defaults(func=sssd_conf)

    ldap_conf_parser = subcommands.add_parser('ldap.conf',
        help="Generate an OpenLDAP ldap.conf configuration for this LDAP server")
    ldap_conf_parser.set_defaults(func=ldap_conf)

    display_parser = subcommands.add_parser('display',
        help="Display generic application parameters for LDAP connection")
    display_parser.set_defaults(func=display)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""Helpers for managing the directory server internal logs.
"""

import copy
import datetime as dt
import json
import glob
import re
import gzip
from lib389.utils import ensure_bytes
from lib389._mapped_object_lint import DSLint
from lib389.lint import (
    DSLOGNOTES0001,  # Unindexed search
    DSLOGNOTES0002,  # Unknown attr in search filter
)

# Because many of these settings can change live, we need to check for certain
# attributes all the time.

MONTH_LOOKUP = {
    'Jan': 1,
    'Feb': 2,
    'Mar': 3,
    'Apr': 4,
    'May': 5,
    'Jun': 6,
    'Jul': 7,
    'Aug': 8,
    'Sep': 9,
    'Oct': 10,
    'Nov': 11,
    'Dec': 12,
}


class DirsrvLog(DSLint):
    """Class of functions to working with the various Directory Server logs
    """
    def __init__(self, dirsrv):
        """Initial class
        @param dirsrv - DirSrv object
        """
        self.dirsrv = dirsrv
        self.log = self.dirsrv.log
        self.prog_timestamp = re.compile(r'\[(?P<day>\d*)\/(?P<month>\w*)\/(?P<year>\d*):(?P<hour>\d*):(?P<minute>\d*):(?P<second>\d*)(.(?P<nanosecond>\d*))+\s(?P<tz>[\+\-]\d{4})')   # noqa
        #  JSON timestamp uses strftime %FT%T --> 2025-02-12T17:00:47.663123181 -0500
        self.prog_json_timestamp = re.compile(r'(?P<year>\d*)-(?P<month>\w*)-(?P<day>\d*)T(?P<hour>\d*):(?P<minute>\d*):(?P<second>\d*)(.(?P<nanosecond>\d*))+\s(?P<tz>[\+\-]\d{4})')   # noqa
        self.prog_datetime = re.compile(r'^(?P<timestamp>\[.*\])')
        self.jsonFormat = False

    def _get_log_path(self):
        """Return the current log file location"""
        raise Exception("Log type not defined.")

    def _get_all_log_paths(self):
        """Return all the log paths"""
        return glob.glob("%s.*-*" % self._get_log_path()) + [self._get_log_path()]

    def readlines_archive(self):
        """
        Returns an array of all the lines in all logs, included rotated logs
        and compressed logs. (gzip)
        Will likely be very slow. Try using match instead.

        @return - an array of all the lines in all logs
        """
        lines = []
        for log in self._get_all_log_paths():
            # Open the log
            if log.endswith(ensure_bytes('.gz')):
                with gzip.open(log, 'r') as lf:
                    lines += lf.readlines()
            else:
                with open(log, 'r') as lf:
                    lines += lf.readlines()
        return lines

    def readlines(self):
        """Returns an array of all the lines in the log.
        Will likely be very slow. Try using match instead.

        @return - an array of all the lines in the log.
        """
        lines = []
        self.lpath = self._get_log_path()
        if self.lpath is not None:
            # Open the log
            with open(self.lpath, 'r') as lf:
                lines = lf.readlines()
        return lines

    def match_archive(self, pattern):
        """Search all the log files, including "zipped" logs
        @param pattern - a regex pattern
        @return - results of the pattern matching
        """
        results = []
        prog = re.compile(pattern)
        for log in self._get_all_log_paths():
            if log.endswith(ensure_bytes('.gz')):
                with gzip.open(log, 'r') as lf:
                    for line in lf:
                        mres = prog.match(line)
                        if mres:
                            results.append(line)
            else:
                with open(log, 'r') as lf:
                    for line in lf:
                        mres = prog.match(line)
                        if mres:
                            results.append(line)
        return results

    def match(self, pattern, after_pattern=None):
        """Search the current log file for the pattern
        @param pattern - a regex pattern
        @param after_pattern - None or a regex pattern
               if set only matches found seeing last occurance
               of after_pattern are returned
        @return - results of the pattern matching
        """
        results = []
        prog = re.compile(pattern)
        if after_pattern:
            aprog = re.compile(after_pattern)
            aprog_hit = False
        else:
            aprog_hit = True
        self.lpath = self._get_log_path()
        if self.lpath is not None:
            with open(self.lpath, 'r') as lf:
                for line in lf:
                    if after_pattern is not None:
                        if aprog.match(line):
                            aprog_hit = True
                            results.clear()
                    mres = prog.match(line)
                    if aprog_hit and mres:
                        results.append(line)
        return results

    def parse_timestamp(self, ts, json_format=False):
        """Parse a logs timestamps and break it down into its individual parts
        @param ts - The timestamp string from a log
        @return - a "datetime" object
        """
        if json_format:
            timedata = self.prog_json_timestamp.match(ts).groupdict()
        else:
            timedata = self.prog_timestamp.match(ts).groupdict()

        # Convert month to an int.
        month = timedata['month']
        if not month.isdigit():
            month = MONTH_LOOKUP[month]
        else:
            month = int(month)

        # Parse timezone offset string (e.g. "+1000" or "-0500") into a timezone
        tz_str = timedata['tz']
        tz_sign = 1 if tz_str[0] == '+' else -1
        tz_hours = int(tz_str[1:3])
        tz_minutes = int(tz_str[3:5])
        tz = dt.timezone(dt.timedelta(hours=tz_sign * tz_hours, minutes=tz_sign * tz_minutes))

        parsed_dt = dt.datetime(
            int(timedata['year']),
            month,
            int(timedata['day']),
            int(timedata['hour']),
            int(timedata['minute']),
            int(timedata['second']),
            tzinfo=tz
        )
        if timedata['nanosecond']:
            parsed_dt = parsed_dt.replace(microsecond=int(timedata['nanosecond']) // 1000)
        return parsed_dt

    def get_time_in_secs(self, log_line):
        """Take the timestamp (not the date) from a DS access log and convert
           it to seconds:

              [25/May/2016:15:24:27.289341875 -0400]...

              JSON format

              2025-02-12T17:00:47.699153015 -0500

            @param log_line - A line of txt from a DS error/access log
            @return - time in seconds
        """

        total = 0

        try:
            log_obj = json.loads(log_line)
            # JSON format
            date_str = log_obj["local_time"]
            time_str = date_str.split('T')[1:]  # splice off the date
            hms = time_str[:8]
            parts = hms.split(':')

        except ValueError:
            # Old format
            index = log_line.index(':') + 1
            hms = log_line[index: index + 8]
            parts = hms.split(':')

        if int(parts[0]):
            total += int(parts[0]) * 3600
        if int(parts[1]):
            total += int(parts[1]) * 60
        total += int(parts[2])

        return total


class DirsrvAccessJSONLog(DirsrvLog):
    """Class for process access logs in JSON format"""
    def __init__(self, dirsrv):
        """Init the class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvAccessJSONLog, self).__init__(dirsrv)
        self.lpath = None

    @classmethod
    def lint_uid(cls):
        return 'logs'

    def _log_get_search_stats(self, conn, op):
        self.lpath = self._get_log_path()
        if self.lpath is not None:
            with open(self.lpath, 'r') as lf:
                for line in lf:
                    line = line.strip()
                    action = json.loads(line)
                    if 'header' in action:
                        # This is the log title, skip it
                        continue
                    if action['operation'] == "SEARCH" and \
                       action['conn_id'] == op and \
                       action['op_id'] == op:
                        return action
        return None

    def _lint_notes(self):
        """
        Check for notes=A (fully unindexed searches), and
        notes=F (unknown attribute in filter)
        """
        for pattern, lint_report in [(".* notes=A", DSLOGNOTES0001), (".* notes=F", DSLOGNOTES0002)]:
            lines = self.match(pattern)
            if len(lines) > 0:
                count = 0
                searches = []
                for line in lines:
                    line = line.strip()
                    action = json.loads(line)
                    if action['operation'] == 'RESULT':
                        # Looks like a valid notes=A/F
                        conn = action['conn_id']
                        op = action['op_id']
                        etime = action['etime']
                        stats = self._log_get_search_stats(conn, op)
                        if stats is not None:
                            timestamp = stats['local_time']
                            base = stats['base_dn']
                            scope = stats['scope']
                            srch_filter = stats['filter']
                            count += 1
                            if lint_report == DSLOGNOTES0001:
                                searches.append(f'\n  [{count}] Unindexed Search\n'
                                                f'      - date:    {timestamp}\n'
                                                f'      - conn/op: {conn}/{op}\n'
                                                f'      - base:    {base}\n'
                                                f'      - scope:   {scope}\n'
                                                f'      - filter:  {srch_filter}\n'
                                                f'      - etime:   {etime}\n')
                            else:
                                searches.append(f'\n  [{count}] Invalid Attribute in Filter\n'
                                                f'      - date:    {timestamp}\n'
                                                f'      - conn/op: {conn}/{op}\n'
                                                f'      - filter:  {srch_filter}\n')
                if len(searches) > 0:
                    report = copy.deepcopy(lint_report)
                    report['items'].append(self._get_log_path())
                    report['detail'] = report['detail'].replace('NUMBER',
                                                                str(count))
                    for srch in searches:
                        report['detail'] += srch
                    report['check'] = 'logs:notes'
                    yield report

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.access_log

    def parse_line(self, line):
        """
        This knows how to break up an access log line into the specific fields.
        @param line - A text line from an access log
        @return - A dictionary of the log parts
        """
        line = line.strip()
        try:
            action = json.loads(line)
        except ValueError:
            # Old format, skip this line
            return None

        # First, pull some well known info out.
        if self.dirsrv.verbose:
            self.log.info("parse_line --> %s ", line)

        action['datetime'] = self.parse_timestamp(action['local_time'],
                                                  json_format=True)

        if self.dirsrv.verbose:
            self.log.info(action)
        return action

    def parse_log(self):
        """
        Take the entire logs and parse it into a list of objects, this can
        handle "json_pretty" format
        """

        json_objects = []
        jobj = ""

        lines = self.readlines()
        for line in lines:
            line = line.rstrip()
            if line == '{':
                jobj = "{"
            elif line == '}':
                jobj += "}"
                json_objects.append(json.loads(jobj))
            else:
                if line[0] == '{' and line[-1] == '}':
                    # Complete json log line
                    json_objects.append(json.loads(line))
                else:
                    # Json pretty - append the line
                    jobj += line.strip()

        return json_objects


class DirsrvAccessLog(DirsrvLog):
    """Class for process access logs"""
    def __init__(self, dirsrv):
        """Init the class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvAccessLog, self).__init__(dirsrv)
        # We precompile our regex for parse_line to make it faster.
        self.prog_m1 = re.compile(r'^(?P<timestamp>\[.*\])\sconn=(?P<conn>\d*)\sop=(?P<op>\d*)\s(?P<action>\w*)\s(?P<rem>.*)')
        self.prog_con = re.compile(r'^(?P<timestamp>\[.*\])\sconn=(?P<conn>\d*)\sfd=(?P<fd>\d*)\sslot=(?P<slot>\d*)\sconnection\sfrom\s(?P<remote>[^\s]*)\sto\s(?P<local>[^\s]*)')
        self.prog_discon = re.compile(r'^(?P<timestamp>\[.*\])\sconn=(?P<conn>\d*)\sop=(?P<op>\d*)\sfd=(?P<fd>\d*)\s(?P<action>closed)\s-\s(?P<status>\w*)')
        # RESULT regex's (based off action.rem)
        self.prog_notes = re.compile(r'err=(?P<err>\d*)\stag=(?P<tag>\d*)\snentries=(?P<nentries>\d*)\setime=(?P<etime>[0-9.]*)\snotes=(?P<notes>\w*)')
        self.prog_repl = re.compile(r'err=(?P<err>\d*)\stag=(?P<tag>\d*)\snentries=(?P<nentries>\d*)\setime=(?P<etime>[0-9.]*)\scsn=(?P<csn>\w*)')
        self.prog_result = re.compile(r'err=(?P<err>\d*)\stag=(?P<tag>\d*)\snentries=(?P<nentries>\d*)\setime=(?P<etime>[0-9.]*)\s(?P<rem>.*)')
        # Lists for each regex type
        self.full_regexs = [self.prog_m1, self.prog_con, self.prog_discon]
        self.result_regexs = [self.prog_notes, self.prog_repl,
                              self.prog_result]

    @classmethod
    def lint_uid(cls):
        return 'logs'

    def _log_get_search_stats(self, conn, op):
        lines = self.match(f".* conn={conn} op={op} SRCH base=.*")
        if len(lines) != 1:
            return None

        quoted_vals = re.findall('"([^"]*)"', lines[0])
        return {
            'base': quoted_vals[0],
            'filter': quoted_vals[1],
            'timestamp': re.findall('[(.*)]', lines[0])[0],
            'scope': lines[0].split(' scope=', 1)[1].split(' ',1)[0]
        }

    def _lint_notes(self):
        """
        Check for notes=A (fully unindexed searches), and
        notes=F (unknown attribute in filter)
        """
        for pattern, lint_report in [(".* notes=A", DSLOGNOTES0001), (".* notes=F", DSLOGNOTES0002)]:
            lines = self.match(pattern)
            if len(lines) > 0:
                count = 0
                searches = []
                for line in lines:
                    if ' RESULT err=' in line:
                        # Looks like a valid notes=A/F
                        conn = line.split(' conn=', 1)[1].split(' ', 1)[0]
                        op = line.split(' op=', 1)[1].split(' ', 1)[0]
                        etime = line.split(' etime=', 1)[1].split(' ', 1)[0]
                        stats = self._log_get_search_stats(conn, op)
                        if stats is not None:
                            timestamp = stats['timestamp']
                            base = stats['base']
                            scope = stats['scope']
                            srch_filter = stats['filter']
                            count += 1
                            if lint_report == DSLOGNOTES0001:
                                searches.append(f'\n  [{count}] Unindexed Search\n'
                                                f'      - date:    {timestamp}\n'
                                                f'      - conn/op: {conn}/{op}\n'
                                                f'      - base:    {base}\n'
                                                f'      - scope:   {scope}\n'
                                                f'      - filter:  {srch_filter}\n'
                                                f'      - etime:   {etime}\n')
                            else:
                                searches.append(f'\n  [{count}] Invalid Attribute in Filter\n'
                                                f'      - date:    {timestamp}\n'
                                                f'      - conn/op: {conn}/{op}\n'
                                                f'      - filter:  {srch_filter}\n')
                if len(searches) > 0:
                    report = copy.deepcopy(lint_report)
                    report['items'].append(self._get_log_path())
                    report['detail'] = report['detail'].replace('NUMBER',
                                                                str(count))
                    for srch in searches:
                        report['detail'] += srch
                    report['check'] = 'logs:notes'
                    yield report

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.access_log

    def parse_line(self, line):
        """
        This knows how to break up an access log line into the specific fields.
        @param line - A text line from an access log
        @return - A dictionary of the log parts
        """
        line = line.strip()
        action = {
            'action': 'CONNECT'
        }
        # First, pull some well known info out.
        if self.dirsrv.verbose:
            self.log.info("--> %s ", line)

        for regex in self.full_regexs:
            result = regex.match(line)
            if result:
                action.update(result.groupdict())
                if regex == self.prog_discon:
                    action['action'] = 'DISCONNECT'
                break

        if action['action'] == 'RESULT':
            for regex in self.result_regexs:
                result = regex.match(action['rem'])
                if result:
                    action.update(result.groupdict())
                    break

        action['datetime'] = self.parse_timestamp(action['timestamp'])

        if self.dirsrv.verbose:
            self.log.info(action)
        return action

    def parse_lines(self, lines):
        """Parse multiple log lines
        @param lines - a list of log lines
        @return - A dictionary of the log parts for each line
        """
        return map(self.parse_line, lines)


class DirsrvErrorLog(DirsrvLog):
    """Directory Server Error log class"""
    def __init__(self, dirsrv):
        """Init the Error log class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvErrorLog, self).__init__(dirsrv)
        self.prog_m1 = re.compile(r'^(?P<timestamp>\[.*\])\s(?P<message>.*)')

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.error_log

    def parse_line(self, line):
        """Parse an errors log line
        @line - a text string from an errors log
        @return - A dictionary of the log parts
        """
        line = line.strip()
        action = self.prog_m1.match(line).groupdict()

        action['datetime'] = self.parse_timestamp(action['timestamp'])
        return action

    def parse_lines(self, lines):
        """Parse multiple lines from an errors log
        @param lines - a list of strings/lines from an errors log
        @return - A dictionary of the log parts for each line
        """
        return map(self.parse_line, lines)


class DirsrvErrorJSONLog(DirsrvLog):
    """Directory Server Error JSON log class"""
    def __init__(self, dirsrv):
        """Init the Error log class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvErrorJSONLog, self).__init__(dirsrv)
        self.jsonFormat = True
        self.lpath = ""

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.error_log

    def readlines(self):
        """Returns an array of all the lines in the log.

        @return - an array of all the lines in the log.
        """
        lines = []
        self.lpath = self._get_log_path()
        if self.lpath is not None:
            # Open the log
            with open(self.lpath, 'r', errors='ignore') as lf:
                lines = lf.readlines()
        return lines

    def parse_line(self, line):
        """Parse a error log line
        @line - a text string from a error log
        @return - A dictionary of the log parts
        """
        line = line.strip()

        try:
            action = json.loads(line)
            if 'header' in action:
                # This is the log title, return it as is
                return action

            action['datetime'] = self.parse_timestamp(action['local_time'],
                                                      json_format=True)
            return action

        except json.decoder.JSONDecodeError:
            # Maybe it's json pretty, regardless we can not parse this single
            # line
            pass

        return None

    def parse_log(self):
        """
        Take the entire logs and parse it into a list of objects, this can
        handle "json_pretty" format
        """

        json_objects = []
        jobj = ""

        lines = self.readlines()
        for line in lines:
            line = line.rstrip()
            if line == '{':
                jobj = "{"
            elif line == '}':
                jobj += "}"
                json_objects.append(json.loads(jobj))
            else:
                if line[0] == '{' and line[-1] == '}':
                    # Complete json log line
                    json_objects.append(json.loads(line))
                else:
                    # Json pretty - append the line
                    jobj += line.strip()

        return json_objects


class DirsrvSecurityLog(DirsrvLog):
    """
    Directory Server Security log class

    Currently this is only written in "json", not "json-pretty"
    """
    def __init__(self, dirsrv):
        """Init the Security log class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvSecurityLog, self).__init__(dirsrv)
        self.jsonFormat = True

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.security_log

    def parse_line(self, line):
        """Parse a Security log line
        @line - a text string from a security log
        @return - A dictionary of the log parts
        """
        line = line.strip()
        action = json.loads(line)
        if 'header' in action:
            # This is the log title, return it as is
            return action
        action['datetime'] = action['date']
        return action

    def parse_lines(self, lines):
        """Parse multiple lines from a security log
        @param lines - a lits of strings/lines from a security log
        @return - A dictionary of the log parts for each line
        """
        return map(self.parse_line, lines)


class DirsrvAuditLog(DirsrvLog):
    """Directory Server Audit log class"""
    def __init__(self, dirsrv):
        """Init the Audit log class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvAuditLog, self).__init__(dirsrv)

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.audit_log

    def parse_line(self, line):
        """Parse an audit log line
        @line - a text string from an audit log
        @return - A dictionary of the log parts
        """
        line = line.strip()
        return line.groupdict()

    def parse_lines(self, lines):
        """Parse multiple lines from an audit log
        @param lines - a list of strings/lines from an audit log
        @return - A dictionary of the log parts for each line
        """
        return map(self.parse_line, lines)


class DirsrvAuditJSONLog(DirsrvLog):
    """Directory Server Audit JSON log class"""
    def __init__(self, dirsrv):
        """Init the Audit log class
        @param dirsrv - A DirSrv object
        """
        super(DirsrvAuditJSONLog, self).__init__(dirsrv)
        self.jsonFormat = True
        self.lpath = ""

    def _get_log_path(self):
        """Return the current log file location"""
        return self.dirsrv.ds_paths.audit_log

    def readlines(self):
        """Returns an array of all the lines in the log. Need to ignore
        encoding errors when dealing with the audit log

        @return - an array of all the lines in the log.
        """
        lines = []
        self.lpath = self._get_log_path()
        if self.lpath is not None:
            # Open the log
            with open(self.lpath, 'r', errors='ignore') as lf:
                lines = lf.readlines()
        return lines

    def parse_line(self, line):
        """Parse a audit log line
        @line - a text string from a audit log
        @return - A dictionary of the log parts
        """
        line = line.strip()
        try:
            action = json.loads(line)
            if 'header' in action:
                # This is the log title, return it as is
                return action
            action['datetime'] = action['gm_time']
            return action
        except json.decoder.JSONDecodeError:
            # Maybe it's json pretty, regardless we can not parse this single
            # line
            pass

        return None

    def parse_log(self):
        """
        Take the entire logs and parse it into a list of objects, this can
        handle "json_pretty" format
        """

        json_objects = []
        jobj = ""

        lines = self.readlines()
        for line in lines:
            line = line.rstrip()
            if line == '{':
                jobj = "{"
            elif line == '}':
                jobj += "}"
                json_objects.append(json.loads(jobj))
            else:
                if line[0] == '{' and line[-1] == '}':
                    # Complete json log line
                    json_objects.append(json.loads(line))
                else:
                    # Json pretty - append the line
                    jobj += line.strip()

        return json_objects



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import subprocess
from lib389.cli_base import CustomHelpFormatter


def enable_cockpit(inst, log, args):
    """
    Enable Cockpit socket
    """

    ENABLE_CMD = ['sudo', 'systemctl', 'enable', '--now', 'cockpit.socket']
    try:
        subprocess.run(ENABLE_CMD)
    except subprocess.CalledProcessError as e:
        raise ValueError(f"Failed to enable cockpit socket.  Error {str(e)}")


def open_firewall(inst, log, args):
    """
    Open the firewall for Cockpit service
    """

    OPEN_CMD = ['sudo', 'firewall-cmd', '--add-service=cockpit', '--permanent']
    if args.zone is not None:
        OPEN_CMD.append(f' --zone={args.zone}')
    try:
        subprocess.run(OPEN_CMD)
    except subprocess.CalledProcessError as e:
        raise ValueError(f'Failed to open firewall for service "cockpit"  Error {str(e)}')


def disable_cockpit(inst, log, args):
    """
    Disable Cockpit socket
    """

    DISABLE_CMD = ['sudo', 'systemctl', 'disable', '--now', 'cockpit.socket']
    try:
        subprocess.run(DISABLE_CMD)
    except subprocess.CalledProcessError as e:
        raise ValueError(f"Failed to disable cockpit socket.  Error {str(e)}")


def close_firewall(inst, log, args):
    """
    Close firewall for Cockpit service
    """

    CLOSE_CMD = ['sudo', 'firewall-cmd', '--remove-service=cockpit', '--permanent']
    try:
        subprocess.run(CLOSE_CMD)
    except subprocess.CalledProcessError as e:
        raise ValueError(f'Failed to remove "cockpit" service from firewall.  Error {str(e)}')


def create_parser(subparsers):
    cockpit_parser = subparsers.add_parser('cockpit', help="Enable the Cockpit interface/UI", formatter_class=CustomHelpFormatter)
    subcommands = cockpit_parser.add_subparsers(help="action")

    # Enable socket
    enable_parser = subcommands.add_parser('enable', help='Enable the Cockpit socket', formatter_class=CustomHelpFormatter)
    enable_parser.set_defaults(func=enable_cockpit)

    # Open firewall
    open_parser = subcommands.add_parser('open-firewall', help='Open the firewall for the "cockpit" service', formatter_class=CustomHelpFormatter)
    open_parser.add_argument('--zone', help="The firewall zone")
    open_parser.set_defaults(func=open_firewall)

    # Disable socket
    disable_parser = subcommands.add_parser('disable', help='Disable the Cockpit socket', formatter_class=CustomHelpFormatter)
    disable_parser.set_defaults(func=disable_cockpit)

    # Close firewall
    close_parser = subcommands.add_parser('close-firewall', help='Remove the "cockpit" service from the firewall settings', formatter_class=CustomHelpFormatter)
    close_parser.set_defaults(func=close_firewall)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import ldap

from lib389.tests.topologies import topology_st
from lib389.utils import logging
from lib389.idm.user import UserAccounts
from lib389._constants import DEFAULT_SUFFIX, SECUREPORT_STANDALONE1

from lib389.config import CertmapLegacy

log = logging.getLogger(__name__)

def test_tls_external(topology_st):

    standalone = topology_st.standalone

    # SETUP TLS
    standalone.stop()
    assert(standalone.nss_ssl.reinit() is True)
    assert(standalone.nss_ssl.create_rsa_ca() is True)
    assert(standalone.nss_ssl.create_rsa_key_and_cert() is True)
    # Create a user
    assert(standalone.nss_ssl.create_rsa_user('testuser') is True)
    # Now get the details of where the key and crt are.
    tls_locs = standalone.nss_ssl.get_rsa_user('testuser')
    #  {'ca': ca_path, 'key': key_path, 'crt': crt_path}

    # Start again
    standalone.start()

    users = UserAccounts(standalone, DEFAULT_SUFFIX)
    user = users.create(properties={
        'uid': 'testuser',
        'cn' : 'testuser',
        'sn' : 'user',
        'uidNumber' : '1000',
        'gidNumber' : '2000',
        'homeDirectory' : '/home/testuser'
    })

    standalone.rsa.create()
    # Set the secure port and nsslapd-security
    standalone.config.set('nsslapd-secureport', '%s' % SECUREPORT_STANDALONE1 )
    standalone.config.set('nsslapd-security', 'on')
    standalone.sslport = SECUREPORT_STANDALONE1
    # Now turn on the certmap.
    cm = CertmapLegacy(standalone)
    certmaps = cm.list()
    certmaps['default']['DNComps'] = ''
    certmaps['default']['FilterComps'] = ['cn']
    certmaps['default']['VerifyCert'] = 'off'
    cm.set(certmaps)

    # Check that EXTERNAL is listed in supported mechns.
    assert(standalone.rootdse.supports_sasl_external())
    # Restart to allow certmaps to be re-read: Note, we CAN NOT use post_open
    # here, it breaks on auth. see lib389/__init__.py
    standalone.restart(post_open=False)

    # Now attempt a bind with TLS external
    conn = standalone.openConnection(saslmethod='EXTERNAL', connOnly=True, certdir=standalone.get_cert_dir(), userkey=tls_locs['key'], usercert=tls_locs['crt'])

    assert(conn.whoami_s() == "dn: uid=testuser,ou=People,dc=example,dc=com")

    # Backup version of the code:
    # ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
    # ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, tls_locs['ca'])
    # ldap.set_option(ldap.OPT_X_TLS_KEYFILE, tls_locs['key'])
    # ldap.set_option(ldap.OPT_X_TLS_CERTFILE, tls_locs['crt'])
    # conn = ldap.initialize(standalone.toLDAPURL())

    # sasl_auth = ldap.sasl.external()
    # conn.sasl_interactive_bind_s("", sasl_auth)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import sys
import os
from enum import Enum, IntEnum
from lib389.properties import *

# current module object
mod = sys.modules[__name__]

(
    SUPPLIER_TYPE,
    HUB_TYPE,
    LEAF_TYPE
) = list(range(3))

INSTALL_LATEST_CONFIG = '999999999'

REPLICA_FLAGS_CON = 0

TTL_DEFAULT_VAL = '86400'

# The structure is convenient for replica promote/demote methods
ReplicaRole = Enum("Replica role", "CONSUMER HUB SUPPLIER STANDALONE")

CONSUMER_REPLICAID = 65535

REPLICA_RDONLY_TYPE = 2  # CONSUMER and HUB
REPLICA_WRONLY_TYPE = 1  # SINGLE and MULTI SUPPLIER
REPLICA_RDWR_TYPE = REPLICA_RDONLY_TYPE | REPLICA_WRONLY_TYPE

REPLICA_FLAGS_RDONLY = 0
REPLICA_FLAGS_WRITE = 1

REPLICA_RUV_UUID = "ffffffff-ffffffff-ffffffff-ffffffff"
REPLICA_RUV_FILTER = ('(&(nsuniqueid=ffffffff-ffffffff-ffffffff-ffffffff)(objectclass=nstombstone))')
REPLICA_OC_TOMBSTONE = "nsTombstone"
REPLICATION_BIND_DN = RA_BINDDN
REPLICATION_BIND_PW = RA_BINDPW
REPLICATION_BIND_METHOD = RA_METHOD
REPLICATION_TRANSPORT = RA_TRANSPORT_PROT
REPLICATION_TIMEOUT = RA_TIMEOUT

BDB_CL_FILENAME = "replication_changelog.db"
LMDB_CL_FILENAME = "replication_changelog.db"

# Attributes that should be masked from logging output
SENSITIVE_ATTRS = ['userpassword',
                   'nsslapd-rootpw',
                   'nsds5replicacredentials',
                   'nsmultiplexorcredentials']

TRANS_STARTTLS = "starttls"
TRANS_SECURE = "secure"
TRANS_NORMAL = "normal"
REPL_TRANS_VALUE = {
            TRANS_STARTTLS: 'TLS',
            TRANS_SECURE: 'SSL',
            TRANS_NORMAL: 'LDAP'
        }

defaultProperties = {
            REPLICATION_BIND_DN: "cn=replication manager,cn=config",
            REPLICATION_BIND_PW: "password",
            REPLICATION_BIND_METHOD: "simple",
            REPLICATION_TRANSPORT: REPL_TRANS_VALUE[TRANS_NORMAL],
            REPLICATION_TIMEOUT: str(120)
        }


CFGSUFFIX = "o=NetscapeRoot"

# Some DN constants
DN_DM = "cn=Directory Manager"
PW_DM = "password"
DN_CONFIG = "cn=config"
DN_LDBM = "cn=ldbm database,cn=plugins,cn=config"
DN_CONFIG_LDBM = "cn=config,cn=ldbm database,cn=plugins,cn=config"
DN_CONFIG_LDBM_BDB = "cn=bdb,cn=config,cn=ldbm database,cn=plugins,cn=config"
DN_CONFIG_LDBM_LMDB = "cn=mdb,cn=config,cn=ldbm database,cn=plugins,cn=config"
DN_USERROOT_LDBM = "cn=userRoot,cn=ldbm database,cn=plugins,cn=config"
DN_SCHEMA = "cn=schema"
DN_MONITOR = "cn=monitor"
DN_MONITOR_SNMP = "cn=snmp,cn=monitor"
DN_MONITOR_MEMBEROF = "cn=MemberOf Plugin,cn=monitor"
DN_MONITOR_LDBM = "cn=monitor,cn=ldbm database,cn=plugins,cn=config"
DN_MONITOR_DATABASE = "cn=database,cn=monitor,cn=ldbm database,cn=plugins,cn=config"
DN_PWDSTORAGE_SCHEMES = "cn=Password Storage Schemes,cn=plugins,cn=config"

CMD_PATH_SETUP_DS = "setup-ds.pl"
CMD_PATH_REMOVE_DS = "remove-ds.pl"
CMD_PATH_SETUP_DS_ADMIN = "setup-ds-admin.pl"
CMD_PATH_REMOVE_DS_ADMIN = "remove-ds-admin.pl"

# State of an DirSrv object
DIRSRV_STATE_INIT = 1
DIRSRV_STATE_ALLOCATED = 2
DIRSRV_STATE_OFFLINE = 3
DIRSRV_STATE_RUNNING = 4
DIRSRV_STATE_ONLINE = 5

# So uh  .... localhost.localdomain doesn't always exist. Stop. Using. It.
# LOCALHOST = "localhost.localdomain"
LOCALHOST_IP = "127.0.0.1"
LOCALHOST = "localhost"
LOCALHOST_SHORT = "localhost"
DEFAULT_PORT = 389
DEFAULT_SECURE_PORT = 636
DEFAULT_SUFFIX = 'dc=example,dc=com'
DEFAULT_SUFFIX_ESCAPED = 'dc\3Dexample\2Cdc\3Dcom'
DEFAULT_DOMAIN = 'example.com'
DEFAULT_BENAME = 'userRoot'  # warning it is case sensitive
DEFAULT_BACKUPDIR = '/tmp'
DEFAULT_INST_HEAD = 'slapd-'
DEFAULT_ENV_HEAD = 'dirsrv-'
DEFAULT_CHANGELOG_NAME = "changelog5"
DEFAULT_CHANGELOG_DB = 'changelogdb'
DEFAULT_DB_LIB = 'mdb'

# CONF_DIR = 'etc/dirsrv'
# ENV_SYSCONFIG_DIR = '/etc/sysconfig'
ENV_LOCAL_DIR = '.dirsrv'

# CONFIG file (<prefix>/etc/sysconfig/dirsrv-* or
# $HOME/.dirsrv/dirsrv-*) keywords
CONF_SERVER_ID = 'SERVER_ID'
CONF_SERVER_DIR = 'SERVER_DIR'
CONF_SERVERBIN_DIR = 'SERVERBIN_DIR'
CONF_CONFIG_DIR = 'CONFIG_DIR'
CONF_INST_DIR = 'INST_DIR'
CONF_RUN_DIR = 'RUN_DIR'
CONF_DS_ROOT = 'DS_ROOT'
CONF_PRODUCT_NAME = 'PRODUCT_NAME'
CONF_LDAPI_ENABLED = 'LDAPI_ENABLED'
CONF_LDAPI_SOCKET = 'LDAPI_SOCKET'
CONF_LDAPI_AUTOBIND = 'LDAPI_AUTOBIND'
CONF_LDAPI_ROOTUSER = 'LDAPI_ROOTUSER'

DN_CONFIG = "cn=config"
DN_PLUGIN = "cn=plugins,%s" % DN_CONFIG
DN_MAPPING_TREE = "cn=mapping tree,%s" % DN_CONFIG
DN_CHANGELOG = "cn=changelog5,%s" % DN_CONFIG
DN_LDBM = "cn=ldbm database,%s" % DN_PLUGIN
DN_CHAIN = "cn=chaining database,%s" % DN_PLUGIN

DN_TASKS = "cn=tasks,%s" % DN_CONFIG
DN_INDEX_TASK = "cn=index,%s" % DN_TASKS
DN_EXPORT_TASK = "cn=export,%s" % DN_TASKS
DN_IMPORT_TASK = "cn=import,%s" % DN_TASKS
DN_BACKUP_TASK = "cn=backup,%s" % DN_TASKS
DN_RESTORE_TASK = "cn=restore,%s" % DN_TASKS
DN_MBO_TASK = "cn=memberOf task,%s" % DN_TASKS
DN_EUUID_TASK = "cn=entryuuid task,%s" % DN_TASKS
DN_TOMB_FIXUP_TASK = "cn=fixup tombstones,%s" % DN_TASKS
DN_FIXUP_LINKED_ATTIBUTES = "cn=fixup linked attributes,%s" % DN_TASKS
DN_AUTOMEMBER_REBUILD_TASK = "cn=automember rebuild membership,%s" % DN_TASKS
DN_AUTOMEMBER_ABORT_REBUILD_TASK = "cn=automember abort rebuild,%s" % DN_TASKS
DN_COMPACTDB_TASK = "cn=compact db,%s" % DN_TASKS
DN_SHADOW_FIXUP_TASKS = "cn=fixup shadow attributes,%s" % DN_TASKS

# Script Constants
LDIF2DB = 'ldif2db'
DB2LDIF = 'db2ldif'
BAK2DB = 'bak2db'
DB2BAK = 'db2bak'
DB2INDEX = 'db2index'
DBSCAN = 'dbscan'

# Task warnings
class TaskWarning(IntEnum):
    WARN_UPGARDE_DN_FORMAT_ALL      = (1 << 0)
    WARN_UPGRADE_DN_FORMAT          = (1 << 1)
    WARN_UPGRADE_DN_FORMAT_SPACE    = (1 << 2)
    WARN_SKIPPED_IMPORT_ENTRY       = (1 << 3)

RDN_REPLICA = "cn=replica"

RETROCL_SUFFIX = "cn=changelog"

##################################
#
# Request Control OIDS
#
##################################
CONTROL_DEREF = '1.3.6.1.4.1.4203.666.5.16'
CONTROL_USE_ONE_BACKEND = '2.16.840.1.113730.3.4.14'
CONTROL_USE_ONE_BACKEND_EXT = '2.16.840.1.113730.3.4.20'

##################################
#
# Plugins
#
##################################

PLUGIN_7_BIT_CHECK = '7-bit check'
PLUGIN_ACCT_POLICY = 'Account Policy Plugin'
PLUGIN_ACCT_USABILITY = 'Account Usability Plugin'
PLUGIN_ACL = 'ACL Plugin'
PLUGIN_ACL_PREOP = 'ACL preoperation'
PLUGIN_ATTR_UNIQUENESS = 'attribute uniqueness'
PLUGIN_AUTOMEMBER = 'Auto Membership Plugin'
PLUGIN_CHAININGDB = 'chaining database'
PLUGIN_COLLATION = 'Internationalization Plugin'
PLUGIN_COS = 'Class of Service'
PLUGIN_DEREF = 'deref'
PLUGIN_DNA = 'Distributed Numeric Assignment Plugin'
PLUGIN_HTTP = 'HTTP Client'
PLUGIN_LINKED_ATTRS = 'Linked Attributes'
PLUGIN_MANAGED_ENTRY = 'Managed Entries'
PLUGIN_MEMBER_OF = 'MemberOf Plugin'
PLUGIN_PAM_PASSTHRU = 'PAM Pass Through Auth'
PLUGIN_PASSTHRU = 'Pass Through Authentication'
PLUGIN_POSIX_WINSYNC = 'Posix Winsync API'
PLUGIN_REFER_INTEGRITY = 'referential integrity postoperation'
PLUGIN_REPL_SYNC = 'Content Synchronization'
PLUGIN_REPLICATION_LEGACY = 'Legacy Replication Plugin'
PLUGIN_REPLICATION = 'Multisupplier Replication Plugin'
PLUGIN_RETRO_CHANGELOG = 'Retro Changelog Plugin'
PLUGIN_ROLES = 'Roles Plugin'
PLUGIN_ROOTDN_ACCESS = 'RootDN Access Control'
PLUGIN_SCHEMA_RELOAD = 'Schema Reload'
PLUGIN_STATECHANGE = 'State Change Plugin'
PLUGIN_USN = 'USN'
PLUGIN_VIEWS = 'Views'
PLUGIN_WHOAMI = 'whoami'
PLUGIN_ADDN = 'addn'


#
# Constants
#
DEFAULT_USER = "dirsrv"
DEFAULT_USERHOME = "/tmp/lib389_home"
DEFAULT_USER_COMMENT = "lib389 DS user"
DATA_DIR = "data"
TMP_DIR = "tmp"
VALGRIND_WRAPPER = "ns-slapd.valgrind"
VALGRIND_LEAK_STR = " blocks are definitely lost in loss record "
VALGRIND_INVALID_STR = " Invalid (free|read|write)"
DISORDERLY_SHUTDOWN = ('Detected Disorderly Shutdown last time Directory '
                       'Server was running, recovering database')
DEFAULT_LMDB_SIZE = '20Gb'

#
# LOG: see https://access.redhat.com/documentation/en-US/Red_Hat_Directory
# _Server/10/html/Administration_Guide/Configuring_Logs.html
# The default log level is 16384
#
# It is legacy constants. Please, use IntEnum version (ErrorLog and AccessLog)
(LOG_TRACE,
 LOG_TRACE_PACKETS,
 LOG_TRACE_HEAVY,
 LOG_CONNECT,
 LOG_PACKET,
 LOG_SEARCH_FILTER,
 LOG_CONFIG_PARSER,
 LOG_ACL,
 LOG_ENTRY_PARSER,
 LOG_HOUSEKEEPING,
 LOG_REPLICA,
 LOG_DEFAULT,
 LOG_CACHE,
 LOG_PLUGIN,
 LOG_MICROSECONDS,
 LOG_ACL_SUMMARY) = [1 << x for x in (list(range(8)) + list(range(11, 19)))]


class ErrorLog(IntEnum):
    (TRACE,
     TRACE_PACKETS,
     TRACE_HEAVY,
     CONNECT,
     PACKET,
     SEARCH_FILTER,
     CONFIG_PARSER,
     ACL,
     ENTRY_PARSER,
     HOUSEKEEPING,
     REPLICA,
     DEFAULT,
     CACHE,
     PLUGIN,
     MICROSECONDS,
     ACL_SUMMARY) = [1 << x for x in (list(range(8)) + list(range(11, 19)))]


class AccessLog(IntEnum):
    NONE = 0
    INTERNAL = 4
    DEFAULT = 256  # Default log level
    ENTRY = 512
    MICROSECONDS = 131072

#
# Constants for individual tests
#
SUFFIX = 'dc=example,dc=com'
PASSWORD = 'password'

# Generate instance parameters on the fly
N=0
port_start = 38901
number_of_instances = 99

# Standalone topology
for i in range(port_start, port_start + number_of_instances):
    N+=1
    setattr(mod, "HOST_STANDALONE{0}".format(N), "LOCALHOST")
    setattr(mod, "PORT_STANDALONE{0}".format(N), i)
    setattr(mod, "SECUREPORT_STANDALONE{0}".format(N), i + 24700)
    setattr(mod, "SERVERID_STANDALONE{0}".format(N), "standalone{0}".format(N))
    setattr(mod, "REPLICAID_STANDALONE_{0}".format(N), 65535)

# For compatibility
HOST_STANDALONE = HOST_STANDALONE1
PORT_STANDALONE = PORT_STANDALONE1
SECUREPORT_STANDALONE = SECUREPORT_STANDALONE1
SERVERID_STANDALONE = SERVERID_STANDALONE1

# Replication topology - suppliers
N=0
port_start+=100
for i in range(port_start, port_start + number_of_instances):
    N+=1
    setattr(mod, "HOST_SUPPLIER_{0}".format(N), "LOCALHOST")
    setattr(mod, "PORT_SUPPLIER_{0}".format(N), i)
    setattr(mod, "SECUREPORT_SUPPLIER_{0}".format(N), i + 24700)
    setattr(mod, "SERVERID_SUPPLIER_{0}".format(N), "supplier{0}".format(N))
    setattr(mod, "REPLICAID_SUPPLIER_{0}".format(N), N)

# Replication topology - hubs
N=0
port_start+=100
for i in range(port_start, port_start + number_of_instances):
    N+=1
    setattr(mod, "HOST_HUB_{0}".format(N), "LOCALHOST")
    setattr(mod, "PORT_HUB_{0}".format(N), i)
    setattr(mod, "SECUREPORT_HUB_{0}".format(N), i + 24700)
    setattr(mod, "SERVERID_HUB_{0}".format(N), "hub{0}".format(N))
    setattr(mod, "REPLICAID_HUB_{0}".format(N), 65535)

# Replication topology - consumers
N=0
port_start+=100
for i in range(port_start, port_start + number_of_instances):
    N+=1
    setattr(mod, "HOST_CONSUMER_{0}".format(N), "LOCALHOST")
    setattr(mod, "PORT_CONSUMER_{0}".format(N), i)
    setattr(mod, "SECUREPORT_CONSUMER_{0}".format(N), i + 24700)
    setattr(mod, "SERVERID_CONSUMER_{0}".format(N), "consumer{0}".format(N))

# Cleanup, we don't need to export that
del N, port_start, number_of_instances

# This is a template
args_instance = {SER_DEPLOYED_DIR: os.environ.get('PREFIX', None),
                 SER_BACKUP_INST_DIR: os.environ.get('BACKUPDIR',
                                                     DEFAULT_BACKUPDIR),
                 SER_ROOT_DN: DN_DM,
                 SER_ROOT_PW: PW_DM,
                 SER_HOST: LOCALHOST,
                 SER_PORT: DEFAULT_PORT,
                 SER_SERVERID_PROP: "template",
                 SER_CREATION_SUFFIX: DEFAULT_SUFFIX}

# Helper for linking dse.ldif values to the parse_config function
args_dse_keys = SER_PROPNAME_TO_ATTRNAME

DSRC_HOME = '~/.dsrc'
DSRC_CONTAINER = '/data/config/container.inf'

CONTAINER_TLS_SERVER_KEY = '/data/tls/server.key'
CONTAINER_TLS_SERVER_CERT = '/data/tls/server.crt'
CONTAINER_TLS_SERVER_CADIR = '/data/tls/ca'
CONTAINER_TLS_PWDFILE = '/data/config/pwdfile.txt'

# Describe what kind of Berkeley Database library is available
BDB_IMPL_STATUS = Enum('BDB_IMPL_STATUS', [
                       'UNKNOWN',   # Unable to discover
                       'STANDARD',  # os libdb rpm is installed
                       'BUNDLED',   # lib389 bundled rpm is installed
                       'READ_ONLY', # Read-only version is available
                       'NONE' ])    # bdb is not usasable

# DB implementation
DB_IMPL_BDB = "bdb"
DB_IMPL_MDB = "mdb"



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObject, DSLdapObjects


class SaslMapping(DSLdapObject):
    """A sasl map providing a link from a sasl user and realm
    to a valid directory server entry.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = [
            'cn',
            'nsSaslMapRegexString',
            'nsSaslMapBaseDNTemplate',
            'nsSaslMapFilterTemplate',
            'nsSaslMapPriority'
        ]

    def __init__(self, instance, dn=None):
        super(SaslMapping, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'

        self._create_objectclasses = [
            'top',
            'nsSaslMapping',
        ]
        self._protected = False


class SaslMappings(DSLdapObjects):
    """DSLdapObjects that represents SaslMappings in the server.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all group entries below
    :type basedn: str
    """

    def __init__(self, instance, dn=None):
        super(SaslMappings, self).__init__(instance)
        self._objectclasses = [
            'nsSaslMapping',
        ]
        self._filterattrs = ['cn']
        self._childobject = SaslMapping
        self._basedn = 'cn=mapping,cn=sasl,cn=config'





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""Tools for creating and managing servers

    uses DirSrv
"""
import sys
import os
import os.path
import base64
import select
import time
import shutil
import subprocess
import tarfile
import urllib
import re
import glob
import pwd
import grp
import logging
import ldap
import shlex
# from .nss_ssl import nss_create_new_database
from threading import Timer
from lib389.paths import Paths
from lib389._constants import *
from lib389.properties import *
from lib389.utils import (
    ensure_bytes,
    ensure_str,
    ds_is_older,
)

__all__ = ['DirSrvTools']

_ds_paths = Paths()

log = logging.getLogger(__name__)

# Private constants
PATH_SETUP_DS_ADMIN = CMD_PATH_SETUP_DS_ADMIN
PATH_SETUP_DS = CMD_PATH_SETUP_DS
PATH_REMOVE_DS = CMD_PATH_REMOVE_DS
PATH_ADM_CONF = "/etc/dirsrv/admin-serv/adm.conf"
GROUPADD = "/usr/sbin/groupadd"
USERADD = "/usr/sbin/useradd"
NOLOGIN = "/sbin/nologin"


def kill_proc(proc, timeout):
    """Kill a process after the timeout is reached
    @param proc - The subprocess process
    @param timeout - timeout in seconds
    """
    timeout["value"] = True
    proc.kill()


def runCmd(cmd, timeout_sec):
    """Run a system command with a timeout
    @param cmd - The full system command
    @param timeout_sec - The timeoput value in seconds
    @return - The result code
    """
    proc = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    timeout = {"value": False}
    timer = Timer(timeout_sec, kill_proc, [proc, timeout])
    timer.start()
    stdout, stderr = proc.communicate()
    timer.cancel()
    return proc.returncode

class DirSrvTools(object):
    """DirSrv mix-in."""

    @staticmethod
    def cgiFake(sroot, verbose, prog, args):
        """Run the local program prog as a CGI using the POST method."""
        content = urllib.parse.urlencode(args)
        length = len(content)
        # setup CGI environment
        env = os.environ.copy()
        env['REQUEST_METHOD'] = "POST"
        env['NETSITE_ROOT'] = sroot
        env['CONTENT_LENGTH'] = str(length)
        progdir = os.path.dirname(prog)
        pipe = subprocess.Popen(prog, cwd=progdir, env=env,
                                stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        child_stdin = pipe.stdin
        child_stdout = pipe.stdout
        child_stdin.write(content)
        child_stdin.close()
        for line in child_stdout:
            if verbose:
                sys.stdout.write(line)
            ary = line.split(":")
            if len(ary) > 1 and ary[0] == 'NMC_Status':
                exitCode = ary[1].strip()
                break
        child_stdout.close()
        osCode = pipe.wait()
        print("%s returned NMC code %s and OS code %s" %
              (prog, exitCode, osCode))
        return exitCode

    @staticmethod
    def cgiPost(host, port, username, password, uri, verbose, secure,
                args=None):
        """Post the request to the admin server.

           Admin server requires authentication, so we use the auth handler
           classes.

            NOTE: the url classes in python use the deprecated
            base64.encodestring() function, which truncates lines,
            causing Apache to give us a 400 Bad Request error for the
            Authentication string.  So, we have to tell
            base64.encodestring() not to truncate."""
        args = args or {}
        prefix = 'http'
        if secure:
            prefix = 'https'
        hostport = host + ":" + port
        # construct our url
        url = '%s://%s:%s%s' % (prefix, host, port, uri)
        # tell base64 not to truncate lines
        savedbinsize = base64.MAXBINSIZE
        base64.MAXBINSIZE = 256
        # create the password manager - we don't care about the realm
        passman = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        # add our password
        passman.add_password(None, hostport, username, password)
        # create the auth handler
        authhandler = urllib.request.HTTPBasicAuthHandler(passman)
        # create our url opener that handles basic auth
        opener = urllib.request.build_opener(authhandler)
        # make admin server think we are the console
        opener.addheaders = [('User-Agent', 'Fedora-Console/1.0')]
        if verbose:
            print("requesting url", url)
            sys.stdout.flush()
        exitCode = 1
        try:
            req = opener.open(url, urllib.parse.urlencode(args))
            for line in req:
                if verbose:
                    print (line)
                ary = line.split(":")
                if len(ary) > 1 and ary[0] == 'NMC_Status':
                    exitCode = ary[1].strip()
                    break
            req.close()
#         except IOError, e:
#             print e
#             print e.code
#             print e.headers
#             raise
        finally:
            # restore binsize
            base64.MAXBINSIZE = savedbinsize
        return exitCode

    @staticmethod
    def _infoInstanceBackupFS(dirsrv):
        """
            Return the information to retrieve the backup file of a given
            instance
            It returns:
                - Directory name containing the backup
                  (e.g. /tmp/slapd-standalone.bck)
                - The pattern of the backup files
                  (e.g. /tmp/slapd-standalone.bck/backup*.tar.gz)
        """
        backup_dir = "%s/slapd-%s.bck" % (dirsrv.backupdir, dirsrv.inst)
        backup_pattern = os.path.join(backup_dir, "backup*.tar.gz")
        return backup_dir, backup_pattern

    @staticmethod
    def clearInstanceBackupFS(dirsrv=None, backup_file=None):
        """
            Remove a backup_file or all backup up of a given instance
        """
        if backup_file:
            if os.path.isfile(backup_file):
                try:
                    os.remove(backup_file)
                except:
                    log.info("clearInstanceBackupFS: fail to remove %s" %
                             backup_file)
                    pass
        elif dirsrv:
            backup_dir, backup_pattern = \
                DirSrvTools._infoInstanceBackupFS(dirsrv)
            list_backup_files = glob.glob(backup_pattern)
            for f in list_backup_files:
                try:
                    os.remove(f)
                except:
                    log.info("clearInstanceBackupFS: fail to remove %s" %
                             backup_file)
                    pass

    @staticmethod
    def checkInstanceBackupFS(dirsrv):
        """
            If it exits a backup file, it returns it
            else it returns None
        """

        backup_dir, backup_pattern = DirSrvTools._infoInstanceBackupFS(dirsrv)
        list_backup_files = glob.glob(backup_pattern)
        if not list_backup_files:
            return None
        else:
            # returns the first found backup
            return list_backup_files[0]

    @staticmethod
    def instanceBackupFS(dirsrv):
        """
            Saves the files of an instance under
                /tmp/slapd-<instance_name>.bck/backup_HHMMSS.tar.gz
            and return the archive file name.
            If it already exists a such file, it assums it is a valid backup
            and returns its name

            dirsrv.sroot: root of the instance  (e.g. /usr/lib64/dirsrv)
            dirsrv.inst: instance name (e.g. standalone for
                         /etc/dirsrv/slapd-standalone)
            dirsrv.confdir: root of the instance config (e.g. /etc/dirsrv)
            dirsrv.dbdir: directory where is stored the database
                          (e.g. /var/lib/dirsrv/slapd-standalone/db)
            dirsrv.changelogdir: directory where is stored the changelog
                                (e.g. /var/lib/dirsrv/slapd-supplier/changelogdb)
        """

        # First check it if already exists a backup file
        backup_dir, backup_pattern = DirSrvTools._infoInstanceBackupFS(dirsrv)
        backup_file = DirSrvTools.checkInstanceBackupFS(dirsrv)
        if backup_file is None:
            if not os.path.exists(backup_dir):
                os.makedirs(backup_dir)
        else:
            return backup_file

        # goes under the directory where the DS is deployed
        listFilesToBackup = []
        here = os.getcwd()
        os.chdir(dirsrv.prefix)
        prefix_pattern = "%s/" % dirsrv.prefix

        # build the list of directories to scan
        instroot = "%s/slapd-%s" % (dirsrv.sroot, dirsrv.inst)
        ldir = [instroot]
        if hasattr(dirsrv, 'confdir'):
            ldir.append(dirsrv.confdir)
        if hasattr(dirsrv, 'dbdir'):
            ldir.append(dirsrv.dbdir)
        if hasattr(dirsrv, 'changelogdir'):
            ldir.append(dirsrv.changelogdir)
        if hasattr(dirsrv, 'errlog'):
            ldir.append(os.path.dirname(dirsrv.errlog))
        if hasattr(dirsrv, 'accesslog') and \
           os.path.dirname(dirsrv.accesslog) not in ldir:
            ldir.append(os.path.dirname(dirsrv.accesslog))

        # now scan the directory list to find the files to backup
        for dirToBackup in ldir:
            for root, dirs, files in os.walk(dirToBackup):
                for file in files:
                    name = os.path.join(root, file)
                    name = re.sub(prefix_pattern, '', name)

                    if os.path.isfile(name):
                        listFilesToBackup.append(name)
                        log.debug("instanceBackupFS add = %s (%s)" %
                                  (name, dirsrv.prefix))

        # create the archive
        name = "backup_%s.tar.gz" % (time.strftime("%m%d%Y_%H%M%S"))
        backup_file = os.path.join(backup_dir, name)
        tar = tarfile.open(backup_file, "w:gz")
        tar.extraction_filter = (lambda member, path: member)

        for name in listFilesToBackup:
            if os.path.isfile(name):
                tar.add(name)
        tar.close()
        log.info("instanceBackupFS: archive done : %s" % backup_file)

        # return to the directory where we were
        os.chdir(here)

        return backup_file

    @staticmethod
    def instanceRestoreFS(dirsrv, backup_file):
        """
        """

        # First check the archive exists
        if backup_file is None:
            log.warning("Unable to restore the instance (missing backup)")
            return 1
        if not os.path.isfile(backup_file):
            log.warning("Unable to restore the instance (%s is not a file)" %
                        backup_file)
            return 1

        #
        # Second do some clean up
        #

        # previous db (it may exists new db files not in the backup)
        log.debug("instanceRestoreFS: remove subtree %s/*" % os.path.dirname(dirsrv.dbdir))
        for root, dirs, files in os.walk(os.path.dirname(dirsrv.dbdir)):
            for d in dirs:
                if d not in ("bak", "ldif"):
                    log.debug(
                        "instanceRestoreFS: before restore remove directory "
                        "%s/%s" % (root, d))
                    shutil.rmtree("%s/%s" % (root, d))

        # previous error/access logs
        log.debug("instanceRestoreFS: remove error logs %s" % dirsrv.errlog)
        for f in glob.glob("%s*" % dirsrv.errlog):
                log.debug("instanceRestoreFS: before restore remove file %s" %
                          (f))
                os.remove(f)
        log.debug("instanceRestoreFS: remove access logs %s" %
                  dirsrv.accesslog)
        for f in glob.glob("%s*" % dirsrv.accesslog):
                log.debug("instanceRestoreFS: before restore remove file %s" %
                          (f))
                os.remove(f)

        # Then restore from the directory where DS was deployed
        here = os.getcwd()
        os.chdir(dirsrv.prefix)

        tar = tarfile.open(backup_file)
        tar.extraction_filter = (lambda member, path: member)
        for member in tar.getmembers():
            if os.path.isfile(member.name):
                #
                # restore only writable files
                # It could be a bad idea and preferably restore all.
                # Now it will be easy to enhance that function.
                if os.access(member.name, os.W_OK):
                    log.debug("instanceRestoreFS: restored %s" % member.name)
                    tar.extract(member.name)
                else:
                    log.debug("instanceRestoreFS: not restored %s "
                              "(no write access)" % member.name)
            else:
                log.debug("instanceRestoreFS: restored %s" % member.name)
                tar.extract(member.name)

        tar.close()

        #
        # Now be safe, triggers a recovery at restart
        #
        guardian_file = os.path.join(dirsrv.dbdir, "guardian")
        if os.path.isfile(guardian_file):
            try:
                log.debug("instanceRestoreFS: remove %s" % guardian_file)
                os.remove(guardian_file)
            except:
                log.warning("instanceRestoreFS: fail to remove %s" %
                            guardian_file)
                pass

        os.chdir(here)

    @staticmethod
    def setupSSL(dirsrv, secport=636, sourcedir=None, secargs=None):
        """configure and setup SSL with a given certificate and
           restart the server.

            See DirSrv.configSSL for the secargs values
        """
        e = dirsrv.configSSL(secport, secargs)
        log.info("entry is %r" % [e])
        dn_config = e.dn
        # get our cert dir
        e_config = dirsrv.getEntry(
            dn_config, ldap.SCOPE_BASE, '(objectclass=*)')
        certdir = e_config.getValue('nsslapd-certdir')
        # have to stop the server before replacing any security files
        DirSrvTools.stop(dirsrv)
        # allow secport for SELinux
        if secport != 636:
            try:
                log.debug("Configuring SELinux on port: %s", str(secport))

                subprocess.check_call(["semanage", "port", "-a", "-t",
                                       "ldap_port_t", "-p", "tcp",
                                       str(secport)])
            except OSError:
                log.debug("Likely SELinux not supported")
                pass
            except subprocess.CalledProcessError:
                log.debug("SELinux fails to configure")
                pass

        # eventually copy security files from source dir to our cert dir
        if sourcedir:
            for ff in ['cert8.db', 'key3.db', 'secmod.db', 'pin.txt',
                       'certmap.conf']:
                srcf = os.path.join(sourcedir, ff)
                destf = os.path.join(certdir, ff)
                # make sure dest is writable so we can copy over it
                try:
                    log.info("Copying security files: %s to %s" %
                             (srcf, destf))
                    mode = os.stat(destf).st_mode
                    newmode = mode | 0o600
                    os.chmod(destf, newmode)
                except Exception as e:
                    print(e)
                    pass  # oh well
                # copy2 will copy the mode too
                shutil.copy2(srcf, destf)

        # now, restart the ds
        DirSrvTools.start(dirsrv, True)

    @staticmethod
    def runInfProg(prog, content, verbose, prefix=None):
        """run a program that takes an .inf style file on stdin"""
        cmd = [prog]
        if verbose:
            cmd.append('-ddd')
        else:
            cmd.extend(['-l', '/dev/null'])
        cmd.extend(['-s', '-f', '-'])
        log.debug("running: %s " % cmd)
        pipe = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        child_stdin = pipe.stdin
        child_stdout = pipe.stdout
        child_stdin.write(ensure_bytes(content))
        child_stdin.close()
        if verbose:
            log.debug("PID %s" % pipe.pid)
        while pipe.poll() is None:
            (rr, wr, xr) = select.select([child_stdout], [], [], 1.0)
            if rr and len(rr) > 0:
                line = rr[0].readline()
                if not line:
                    break
                if verbose:
                    sys.stdout.write(ensure_str(line))
            elif verbose:
                print("timed out waiting to read from pid %s : %s " % (pipe.pid, cmd))
        child_stdout.close()
        exitCode = pipe.wait()
        # if verbose:
        log.debug("%s returned exit code %s" % (prog, exitCode))
        return exitCode

    @staticmethod
    def removeInstance(dirsrv):
        """run the remove instance command"""
        prog = os.path.join(_ds_paths.sbin_dir, 'dsctl')
        try:
            cmd = "%s slapd-%s remove --do-it" % (prog, dirsrv.serverid)
            log.info('Running: {}'.format(" ".join(cmd)))
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError as e:
            log.debug('DS has failed to remove the instance {}, '
                      'output is: {}'.format(dirsrv.serverid, e.output))

    @staticmethod
    def _offlineDirsrv(args):
        '''
            Function to allocate an offline DirSrv instance.
            This instance is not initialized with the Directory instance
            (__localinit__() and __add_brookers__() are not called)
            The properties set are:
                instance.host
                instance.port
                instance.serverid
                instance.inst
                instance.prefix
                instance.backup
        '''
        from lib389 import DirSrv
        instance = DirSrv(verbose=True)
        instance.allocate(args)

        return instance

    @staticmethod
    def existsBackup(args):
        '''
            If the backup of the instance exists, it returns it.
            Else None
        '''
        instance = DirSrvTools._offlineDirsrv(args)
        return DirSrvTools.checkInstanceBackupFS(instance)

    @staticmethod
    def existsInstance(args):
        '''
            Check if an instance exists.
            It checks if the following directories/files exist:
                <confdir>/slapd-<name>
                <errlog>
            If it exists it returns a DirSrv instance NOT initialized,
            else None
        '''
        instance = DirSrvTools._offlineDirsrv(args)
        dirname = _ds_paths.config_dir.format(instance_name=instance.serverid)
        errorlog = _ds_paths.error_log.format(instance_name=instance.serverid)
        sroot = os.path.join(instance.prefix, "lib/dirsrv")
        if os.path.isdir(dirname) and \
           os.path.isfile(errorlog) and \
           os.path.isdir(sroot):
            instance.sroot = sroot
            instance.errlog = errorlog
            return instance

        return None

    @staticmethod
    def makeGroup(group=DEFAULT_USER):
        """
        Create a system group
        """
        try:
            grp.getgrnam(group)
            log.debug("OK group %s exists" % group)
        except KeyError:
            log.debug("Adding group %s" % group)
            cmd = [GROUPADD, '-r', group]
            subprocess.Popen(cmd)

    @staticmethod
    def makeUser(user=DEFAULT_USER, group=DEFAULT_USER, home=DEFAULT_USERHOME):
        """
        Create a system user
        """
        try:
            pwd.getpwnam(user)
            log.debug("OK user %s exists" % user)
        except KeyError:
            log.debug("Adding user %s" % user)
            cmd = [USERADD, '-g', group, '-c', DEFAULT_USER_COMMENT, '-r',
                   '-d', home, '-s', NOLOGIN, user]
            subprocess.Popen(cmd)

    @staticmethod
    def lib389User(user=DEFAULT_USER):
        """
        Create the lib389 user/group
        """
        DirSrvTools.makeGroup(group=user)
        time.sleep(1)  # Need a little time for the group to get fully created
        DirSrvTools.makeUser(user=user, group=user, home=DEFAULT_USERHOME)

    @staticmethod
    def searchHostsFile(expectedHost, ipPattern=None):
        """
        Search the hosts file
        """
        hostFile = '/etc/hosts'

        with open(hostFile, 'r') as hostfp:
            # The with statement will automatically close the file after use

            try:
                for line in hostfp.readlines():
                    if ipPattern is None:
                        words = line.split()
                        if len(words) >= 2 and words[1] == expectedHost:
                            return True
                    else:
                        if line.find(ipPattern) >= 0:
                            words = line.split()
                            # We just want to make sure it's in there somewhere
                            if expectedHost in words:
                                return True
            except AssertionError:
                raise AssertionError(
                    "Error: %s should contain '%s' host for %s" %
                    ('/etc/hosts', expectedHost, ipPattern))
            raise AssertionError(
                "Error: /etc/hosts does not contain '%s' as a host for %s"
                % (expectedHost, ipPattern))

    @staticmethod
    def getLocalhost():
        """Get the first host value after 127.0.0.1
        from /etc/hosts file
        """

        with open('/etc/hosts', 'r') as f:
            for line in f.readlines():
                if line.startswith('127.0.0.1'):
                    localhost = line.split()[1]
                    return localhost
        return None

    @staticmethod
    def testLocalhost():
        '''
        Checks that the 127.0.0.1 is resolved as localhost.localdomain
        This is required by DSUtil.pm:checkHostname else setup-ds.pl fails
        '''
        # One day I'll make this ::1. Ipv6 is the future. The future is now.
        loopbackIpPattern = '127.0.0.1'
        DirSrvTools.searchHostsFile(LOCALHOST, loopbackIpPattern)
        DirSrvTools.searchHostsFile(LOCALHOST_SHORT, loopbackIpPattern)

    @staticmethod
    def runUpgrade(prefix, online=True):
        '''
        Run "setup-ds.pl --update"  We simply pass in one DirSrv isntance, and
        this will update all the instances that are in this prefix.  For the
        update to work we must fix/adjust the permissions of the scripts in:

            /prefix/lib[64]/dirsrv/slapd-INSTANCE/
        '''
        if ds_is_older('1.4.0'):
            libdir = os.path.join(_ds_paths.lib_dir, 'dirsrv')

            # Gather all the instances so we can adjust the permissions, otherwise
            servers = []
            path = os.path.join(_ds_paths.sysconf_dir, 'dirsrv')
            for files in os.listdir(path):
                if files.startswith('slapd-') and not files.endswith('.removed'):
                    servers.append(os.path.join(libdir, files))

            if len(servers) == 0:
                # This should not happen
                log.fatal('runUpgrade: no servers found!')
                assert False

            '''
            The setup script calls things like /lib/dirsrv/slapd-instance/db2bak,
            etc, and when we run the setup perl script it gets permission denied
            as the default permissions are 750.  Adjust the permissions to 755.
            '''
            for instance in servers:
                for files in os.listdir(instance):
                    os.chmod(os.path.join(instance, files), 755)

            # Run the "upgrade"
            try:
                prog = os.path.join(_ds_paths.sbin_dir, PATH_SETUP_DS)
                process = subprocess.Popen([prog, '--update'], shell=False,
                                       stdin=subprocess.PIPE)
                # Answer the interactive questions, as "--update" currently does
                # not work with INF files
                process.stdin.write(b'yes\n')
                if(online):
                    process.stdin.write(b'online\n')
                    for x in servers:
                        process.stdin.write(ensure_bytes(DN_DM + '\n'))
                        process.stdin.write(ensure_bytes(PW_DM + '\n'))
                else:
                    process.stdin.write(b'offline\n')
                process.stdin.close()
                process.wait()
                if process.returncode != 0:
                    log.fatal('runUpgrade failed!  Error: %s ' % process.returncode)
                    assert(False)
            except:
                log.fatal('runUpgrade failed!')
                raise
        else:
            pass

    @staticmethod
    def searchFile(filename, pattern):
        """
        Search file (or files using a wildcard) for the pattern
        """
        found = False

        file_list = glob.glob(filename)
        for file_name in file_list:
            try:
                myfile = open(file_name)
                for line in myfile:
                    result = re.search(pattern, line)
                    if result:
                        found = True
                        break
                myfile.close()
                if found:
                    break
            except IOError as e:
                log.error('Problem opening/searching file ' +
                          '(%s): I/O error(%d): %s' %
                          (file_name, e.errno, e.strerror))
        return found


class MockDirSrv(object):
    """
    Mock DirSrv Object
    """
    host = 'localhost'
    port = 22389
    sslport = 0

    def __init__(self, dict_=None):
        """
        Init object
        """
        if dict_:
            self.host = dict_['host']
            self.port = dict_['port']
            if 'sslport' in dict_:
                self.sslport = dict_['sslport']

    def __str__(self):
        """
        Return ldap URL
        """
        if self.sslport:
            return 'ldaps://%s:%s' % (self.host, self.sslport)
        else:
            return 'ldap://%s:%s' % (self.host, self.port)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObject

# Collection of generic LDAP objects/entries

class DSContainer(DSLdapObject):
    """Just a basic container entry

    :param instance: DirSrv instance
    :type instance: DirSrv
    :param dn: The dn of the entry
    :type dn: str
    """

    def __init__(self, instance, dn):
        super(DSContainer, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._create_objectclasses = [
            'top',
            'nsContainer',
        ]
        self._protected = True



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

# Test the cli tools from the dsadm command for correct behaviour.

import pytest
from lib389.cli_ctl.instance import instance_list
from lib389 import DirSrv
from lib389.cli_base import LogCapture


def test_instance_list():
    lc = LogCapture()
    inst = DirSrv()
    instance_list(inst, lc.log, None)
    assert(lc.contains("No instances of Directory Server") or lc.contains("instance: "))
    # Now assert the logs in the capture.

# Need a fixture to install an instance.

# Test start

# Test stop





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest
import ldap

from lib389._constants import DEFAULT_SUFFIX, INSTALL_LATEST_CONFIG

from lib389.cli_conf.backend import backend_create
from lib389.cli_idm.initialise import initialise
from lib389.cli_idm.user import get, create, delete, status, lock, unlock

from lib389.cli_base import LogCapture, FakeArgs
from lib389.tests.cli import topology

from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")

# Topology is pulled from __init__.py
def test_user_tasks(topology):
    be_args = FakeArgs()

    be_args.be_name = 'userRoot'
    be_args.suffix = DEFAULT_SUFFIX
    be_args.parent_suffix = None
    be_args.create_entries = False
    backend_create(topology.standalone, None, topology.logcap.log, be_args)

    # And add the skeleton objects.
    init_args = FakeArgs()
    init_args.version = INSTALL_LATEST_CONFIG
    initialise(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, init_args)

    # First check that our test user isn't there:
    topology.logcap.flush()
    u_args = FakeArgs()
    u_args.selector = 'testuser'
    u_args.json = False
    with pytest.raises(ldap.NO_SUCH_OBJECT):
        get(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)

    # Create the user
    topology.logcap.flush()
    u_args.uid = 'testuser'
    # u_args.sn = 'testuser'
    u_args.cn = 'Test User'
    u_args.displayName = 'Test User'
    u_args.homeDirectory = '/home/testuser'
    u_args.uidNumber = '5000'
    u_args.gidNumber = '5000'
    create(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)

    assert(topology.logcap.contains("Successfully created testuser"))
    # Assert they exist
    topology.logcap.flush()
    get(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)
    assert(topology.logcap.contains('dn: uid=testuser,ou=people,dc=example,dc=com'))

    # Reset the password

    # Lock the account, check status
    topology.logcap.flush()
    lock(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)
    assert(topology.logcap.contains('locked'))

    topology.logcap.flush()
    status(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)
    assert(topology.logcap.contains('locked: True'))

    # Unlock check status
    topology.logcap.flush()
    unlock(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)
    assert(topology.logcap.contains('unlocked'))

    topology.logcap.flush()
    status(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)
    assert(topology.logcap.contains('locked: False'))

    # Enroll a dummy cert

    # Enroll a dummy sshkey

    # Delete it
    topology.logcap.flush()
    u_args.dn = 'uid=testuser,ou=people,dc=example,dc=com'
    delete(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args, warn=False)
    assert(topology.logcap.contains('Successfully deleted uid=testuser,ou=people,dc=example,dc=com'))



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import ldap
from lib389._mapped_object import DSLdapObject, DSLdapObjects, _gen_filter
from lib389.utils import is_a_dn


class ConflictEntry(DSLdapObject):
    """A replication conflict entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: The DN of the conflict entry
    :type dn: str
    """
    def __init__(self, instance, dn=None):
        super(ConflictEntry, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._create_objectclasses = ['ldapsubentry']
        self._protected = False
        self._object_filter = '(objectclass=ldapsubentry)'

    def convert(self, new_rdn):
        """Convert conflict entry to a valid entry, but we need to
        give the conflict entry a new rdn since we are not replacing
        the existing valid counterpart entry.
        """

        if not is_a_dn(new_rdn):
            raise ValueError("The new RDN (" + new_rdn + ") is not a valid DN")

        # Get the conflict entry info
        conflict_value = self.get_attr_val_utf8('nsds5ReplConflict')
        entry_dn = conflict_value.split(' ', 2)[2]
        entry_rdn = ldap.explode_dn(entry_dn, 1)[0]
        rdn_attr = entry_dn.split('=', 1)[0]

        # Rename conflict entry
        self.rename(new_rdn, deloldrdn=False)

        # Cleanup entry
        self.remove(rdn_attr, entry_rdn)
        if self.present('objectclass', 'ldapsubentry'):
            self.remove('objectclass', 'ldapsubentry')
        self.remove_all('nsds5ReplConflict')

    def swap(self):
        """Make the conflict entry the real valid entry.  Delete old valid entry,
        and rename the conflict
        """

        # Get the conflict entry info
        conflict_value = self.get_attr_val_utf8('nsds5ReplConflict')
        entry_dn = conflict_value.split(' ', 2)[2]
        entry_rdn = ldap.explode_dn(entry_dn, 1)[0]

        # Gather the RDN details
        rdn_attr = entry_dn.split('=', 1)[0]
        new_rdn = "{}={}".format(rdn_attr, entry_rdn)
        tmp_rdn = new_rdn + 'tmp'

        # Delete valid entry and its children (to be replaced by conflict entry)
        original_entry = DSLdapObject(self._instance, dn=entry_dn)
        original_entry._protected = False
        filterstr = "(|(objectclass=*)(objectclass=ldapsubentry))"
        ents = self._instance.search_s(original_entry._dn, ldap.SCOPE_SUBTREE, filterstr, escapehatch='i am sure')
        for ent in sorted(ents, key=lambda e: len(e.dn), reverse=True):
            self._instance.delete_ext_s(ent.dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')

        # Rename conflict entry to tmp rdn so we can clean up the rdn attr
        self.rename(tmp_rdn, deloldrdn=False)

        # Cleanup entry
        self.remove(rdn_attr, entry_rdn)
        if self.present('objectclass', 'ldapsubentry'):
            self.remove('objectclass', 'ldapsubentry')
        self.remove_all('nsds5ReplConflict')

        # Rename to the final/correct rdn
        self.rename(new_rdn, deloldrdn=True)

    def get_valid_entry(self):
        """Get the conflict entry's valid counterpart entry
        """
        # Get the conflict entry info
        conflict_value = self.get_attr_val_utf8('nsds5ReplConflict')
        entry_dn = conflict_value.split(' ', 2)[2]

        # Get the valid entry
        return DSLdapObject(self._instance, dn=entry_dn)


class ConflictEntries(DSLdapObjects):
    """Represents the set of tombstone objects that may exist on
    this replica. Tombstones are locally generated, so they are
    unique to individual suppliers, and may or may not correlate
    to tombstones on other suppliers.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Tree to search for tombstones in
    :type basedn: str
    """
    def __init__(self, instance, basedn):
        super(ConflictEntries, self).__init__(instance)
        self._objectclasses = ['ldapsubentry']
        # Try some common ones ....
        self._filterattrs = ['nsds5replconflict', 'objectclass']
        self._childobject = ConflictEntry
        self._basedn = basedn

    def _get_objectclass_filter(self):
        return "(&(objectclass=ldapsubentry)(nsds5replconflict=*))"


class GlueEntry(DSLdapObject):
    """A replication glue entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: The DN of the conflict entry
    :type dn: str
    """
    def __init__(self, instance, dn=None):
        super(GlueEntry, self).__init__(instance, dn)
        self._rdn_attribute = ''
        self._create_objectclasses = ['glue']
        self._protected = False
        self._object_filter = '(objectclass=glue)'

    def convert(self):
        """Convert entry into real entry
        """
        self.remove_all('nsds5replconflict')
        self.remove('objectclass', 'glue')

    def delete_all(self):
        """Remove glue entry and its children.  Depending on the situation the URP
        mechanism can turn the parent glue entry into a tombstone before we get
        a chance to delete it.  This results in a NO_SUCH_OBJECT exception
        """
        delete_count = 0
        filterstr = "(|(objectclass=*)(objectclass=ldapsubentry))"
        ents = self._instance.search_s(self._dn, ldap.SCOPE_SUBTREE, filterstr, escapehatch='i am sure')
        for ent in sorted(ents, key=lambda e: len(e.dn), reverse=True):
            try:
                self._instance.delete_ext_s(ent.dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
                delete_count += 1
            except ldap.NO_SUCH_OBJECT as e:
                if len(ents) > 0 and delete_count == (len(ents) - 1):
                    # This is the parent glue entry - it was removed by URP
                    pass
                else:
                    raise e


class GlueEntries(DSLdapObjects):
    """Represents the set of glue entries that may exist on
    this replica.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Tree to search for tombstones in
    :type basedn: str
    """
    def __init__(self, instance, basedn):
        super(GlueEntries, self).__init__(instance)
        self._objectclasses = ['glue']
        # Try some common ones ....
        self._filterattrs = ['nsds5replconflict', 'objectclass']
        self._childobject = GlueEntry
        self._basedn = basedn

    def _get_objectclass_filter(self):
        return _gen_filter(['objectclass'], ['glue'])



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import os
from lib389.nss_ssl import NssSsl, CERT_NAME, CA_NAME
from lib389.cli_base import _warn, CustomHelpFormatter


def show_servercert(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    nickname = args.nickname or CERT_NAME
    log.info(tls.display_cert_details(nickname))


def list_client_cas(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    # This turns an array of [('CA', 'C,,')]
    for c in tls.list_client_ca_certs():
        log.info(c[0])


def list_cas(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    # Returns a list of cert dicts, eg {'cn': 'nickname', etc}
    for c in tls.list_ca_certs():
        log.info(c['cn'])


def show_cert(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    nickname = args.nickname or CERT_NAME
    log.info(tls.display_cert_details(nickname))


def import_client_ca(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    cert_path = args.cert_path
    nickname = args.nickname or CERT_NAME
    if nickname.lower() == CERT_NAME.lower() or nickname.lower() == CA_NAME.lower():
        log.error("You may not import a CA with the nickname %s or %s" % (CERT_NAME, CA_NAME))
        return
    tls.add_cert(nickname=nickname, cert_file=cert_path)
    tls.edit_cert_trust(nickname, "T,,")


def import_ca(inst, log, args):
    if not os.path.isfile(args.cert_path):
        raise ValueError(f'Certificate file "{args.cert_path}" does not exist')

    tls = NssSsl(dirsrv=inst)
    tls.add_ca_cert_bundle(args.cert_path, args.nickname)


def import_key_cert_pair(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    key_path = args.key_path
    cert_path = args.cert_path
    nickname = args.nickname or CERT_NAME
    tls.add_server_key_and_cert(key_path, cert_path, nickname)


def generate_key_csr(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    alt_names = args.alt_names
    subject = args.subject
    nickname = args.nickname or CERT_NAME
    out_path = tls.create_rsa_key_and_csr(alt_names, subject, name=nickname)
    log.info(out_path)


def import_server_cert(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    cert_path = args.cert_path
    nickname = args.nickname or CERT_NAME
    tls.import_rsa_crt(crt=cert_path, name=nickname)


def remove_cert(inst, log, args, warn=True):
    tls = NssSsl(dirsrv=inst)
    nickname = args.nickname
    if warn:
        _warn(nickname, msg="Deleting certificate %s" % nickname)
    tls.del_cert(nickname)


def export_cert(inst, log, args):
    tls = NssSsl(dirsrv=inst)
    nickname = args.nickname
    der_format = False
    output_file = None
    if args.binary_format:
        der_format = True
    if args.output_file is not None:
        output_file = args.output_file
    tls.export_cert(nickname, output_file, der_format)


def create_parser(subparsers):
    tls_parser = subparsers.add_parser('tls', help="Manage TLS certificates", formatter_class=CustomHelpFormatter)

    subcommands = tls_parser.add_subparsers(help='action')

    list_ca_parser = subcommands.add_parser('list-ca', help='list server certificate authorities including intermediates', formatter_class=CustomHelpFormatter)
    list_ca_parser.set_defaults(func=list_cas)

    list_client_ca_parser = subcommands.add_parser('list-client-ca', help='list client certificate authorities including intermediates', formatter_class=CustomHelpFormatter)
    list_client_ca_parser.set_defaults(func=list_client_cas)

    show_servercert_parser = subcommands.add_parser('show-server-cert', help='Show the active server certificate that clients will see and verify', formatter_class=CustomHelpFormatter)
    show_servercert_parser.add_argument('--nickname', help="The nickname (friendly name) of the certificate to display if different than 'Server-Cert'")
    show_servercert_parser.set_defaults(func=show_servercert)

    show_cert_parser = subcommands.add_parser('show-cert', help='Show a certificate\'s details referenced by it\'s nickname. This is analogous to certutil -L -d <path> -n <nickname>', formatter_class=CustomHelpFormatter)
    show_cert_parser.add_argument('nickname', help="The nickname (friendly name) of the certificate to display")
    show_cert_parser.set_defaults(func=show_cert)

    generate_server_cert_csr_parser = subcommands.add_parser(
        'generate-server-cert-csr',
        help="Generate a Server-Cert certificate signing request - the csr is then submitted to a CA for verification, and when signed you import with import-ca and import-server-cert",
        formatter_class=CustomHelpFormatter
    )
    generate_server_cert_csr_parser.add_argument('--subject', '-s',
        default=None,
        help="Certificate Subject field to use")
    generate_server_cert_csr_parser.add_argument('alt_names', nargs='*',
        help="Certificate requests subject alternative names. These are auto-detected if not provided")
    generate_server_cert_csr_parser.add_argument('--nickname', help="The name of the certificate if different than 'Server-Cert'")
    generate_server_cert_csr_parser.set_defaults(func=generate_key_csr)

    import_client_ca_parser = subcommands.add_parser(
        'import-client-ca',
        help="Import a CA trusted to issue user (client) certificates. This is part of how client certificate authentication functions.",
        formatter_class=CustomHelpFormatter
    )
    import_client_ca_parser.add_argument('cert_path',
        help="The path to the x509 cert to import as a client trust root")
    import_client_ca_parser.add_argument('nickname', help="The name of the certificate once imported")
    import_client_ca_parser.set_defaults(func=import_client_ca)

    import_ca_parser = subcommands.add_parser(
        'import-ca',
        help="Import a CA or intermediate CA for signing this servers certificates (aka Server-Cert). "
             "You should import all the CA's in the chain as required.  PEM bundles are accepted",
        formatter_class=CustomHelpFormatter
    )
    import_ca_parser.add_argument('cert_path',
        help="The path to the x509 cert to import as a server CA")
    import_ca_parser.add_argument('nickname', nargs='+', help="The name of the certificate once imported")
    import_ca_parser.set_defaults(func=import_ca)

    import_server_cert_parser = subcommands.add_parser(
        'import-server-cert',
        help="Import a new Server-Cert after the csr has been signed from a CA.",
        formatter_class=CustomHelpFormatter
    )
    import_server_cert_parser.add_argument('cert_path',
        help="The path to the x509 cert to import as Server-Cert")
    import_server_cert_parser.add_argument('--nickname', help="The name of the certificate if different than 'Server-Cert'")
    import_server_cert_parser.set_defaults(func=import_server_cert)

    import_server_key_cert_parser = subcommands.add_parser(
        'import-server-key-cert',
        help="Import a new key and Server-Cert after having been signed from a CA. This is used if you have an external csr tool or a service like lets encrypt that generates PEM keys externally.",
        formatter_class=CustomHelpFormatter
    )
    import_server_key_cert_parser.add_argument('cert_path',
        help="The path to the x509 cert to import as Server-Cert")
    import_server_key_cert_parser.add_argument('key_path',
        help="The path to the x509 key to import associated to Server-Cert")
    import_server_key_cert_parser.add_argument('--nickname', help="The name of the certificate if different than 'Server-Cert'")
    import_server_key_cert_parser.set_defaults(func=import_key_cert_pair)

    remove_cert_parser = subcommands.add_parser(
        'remove-cert',
        help="Delete a certificate from this database. This will remove it from acting as a CA, a client CA or the Server-Cert role.",
        formatter_class=CustomHelpFormatter
    )
    remove_cert_parser.add_argument('nickname', help="The name of the certificate to delete")
    remove_cert_parser.set_defaults(func=remove_cert)

    export_cert_parser = subcommands.add_parser(
        'export-cert',
        help="Export a certificate to PEM or DER/Binary format.  PEM format is the default",
        formatter_class=CustomHelpFormatter
    )
    export_cert_parser.add_argument('nickname', help="The name of the certificate to export")
    export_cert_parser.add_argument('--binary-format', action='store_true',
                                    help="Export certificate in DER/binary format")
    export_cert_parser.add_argument('--output-file',
                                    help='The name for the exported certificate.  Default name is the certificate '
                                         'nickname with an extension of ".pem" or ".crt"')
    export_cert_parser.set_defaults(func=export_cert)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldif
from lib389._entry import Entry
from lib389.utils import normalizeDN

__all__ = ['LDIFConn']


class LDIFConn(ldif.LDIFParser):
    def __init__(
        self,
        input_file,
        ignored_attr_types=None, max_entries=0, process_url_schemes=None
    ):
        """
        See LDIFParser.__init__()

        Additional Parameters:
        all_records
        List instance for storing parsed records
        """
        self.dndict = {}  # maps dn to Entry
        self.dnlist = []  # contains entries in order read
        myfile = input_file
        if isinstance(input_file, str):
            myfile = open(input_file, "r")
        ldif.LDIFParser.__init__(self, myfile, ignored_attr_types,
                                 max_entries, process_url_schemes)
        self.parse()
        if isinstance(input_file, str):
            myfile.close()

    def handle(self, dn, entry):
        """
        Append single record to dictionary of all records.
        """
        if not dn:
            dn = ''
        newentry = Entry((dn, entry))
        self.dndict[normalizeDN(dn)] = newentry
        self.dnlist.append(newentry)

    def get(self, dn):
        ndn = normalizeDN(dn)
        return self.dndict.get(ndn, Entry(None))



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.idm.account import Account
from lib389._constants import DN_DM, PW_DM


class DirectoryManager(Account):
    """
    The directory manager. This is a convinence class to help with rebinds
    to the same server, as well as some other DM related specific tasks.
    """

    def __init__(self, instance, dn=DN_DM):
        """The Directory Manager instance. Useful for binding in tests.

        :param instance: An instance
        :type instance: lib389.DirSrv
        :param dn: Entry DN
        :type dn: str
        """
        super(DirectoryManager, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = []
        self._create_objectclasses = None
        self._protected = True

    def change_password(self, new_password):
        if new_password == "":
            raise ValueError("You can not set the Directory Manager password to nothing")
        self._instance.config.set('nsslapd-rootpw', new_password)

    def bind(self, password=PW_DM, *args, **kwargs):
        """Bind as the Directory Manager. We have a default test password
        that can be overriden.

        :param password: The password to bind as for Directory Manager
        :type password: str
        :returns: A new connection bound as directory manager.
        """
        return super(DirectoryManager, self).bind(password, *args, **kwargs)

    def rebind(self, password=PW_DM):
        """Rebind on the same connection
        :param password: Directory Manager password
        :type password: str
        """
        self._instance.simple_bind_s(self.dn, password, escapehatch='i am sure')



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import os
import re
import copy
import subprocess
from lib389._mapped_object_lint import DSLint
from lib389 import pid_from_file
from lib389.lint import DSTHPLE0001
from lib389.utils import ensure_str

class Tunables(DSLint):
    """A class for working with system tunables
    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance):
        self._instance = instance
        self.pid = str(pid_from_file(instance.ds_paths.pid_file))


    @classmethod
    def lint_uid(cls):
        return 'tunables'


    def _lint_thp(self):
        """Check if THP is enabled"""
        def systemwide_thp_enabled() -> bool:
            thp_path = '/sys/kernel/mm/transparent_hugepage'
            thp_enabled_path = os.path.join(thp_path, "enabled")
            thp_status_pattern = r"(.*\[always\].*)|(.*\[madvise\].*)"
            if os.path.exists(thp_enabled_path):
                with open(thp_enabled_path, 'r') as f:
                    thp_status = f.read().strip()
                    match = re.match(thp_status_pattern, thp_status)
                    return match is not None


        def instance_thp_enabled(server_running: bool) -> bool:
            pid_status_path = f"/proc/{self.pid}/status"

            if server_running:
                with open(pid_status_path, 'r') as pid_status:
                    pid_status_content = pid_status.read()
                    thp_line = None
                    for line in pid_status_content.split('\n'):
                        if 'THP_enabled' in line:
                            thp_line = line
                            break
                    if thp_line is not None:
                        thp_value = int(thp_line.split()[1])
                        return bool(thp_value)
            else:
                # Get systemctl data for THP from systemd
                systemctl_data = subprocess.check_output(
                    ['systemctl', 'show', '-P', 'Environment',
                     'dirsrv@%s' % self._instance.serverid])

                if 'THP_DISABLE=1' in ensure_str(systemctl_data):
                    return False
                else:
                    return True

        if instance_thp_enabled(self._instance.status()) and systemwide_thp_enabled():
            report = copy.deepcopy(DSTHPLE0001)
            report['check'] = 'tunables:transparent_huge_pages'
            yield report



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import pytest
import logging
import ldap

from lib389._constants import *
from lib389.monitor import MonitorBackend
from lib389.mappingTree import MappingTrees
from lib389.index import Indexes
from lib389.backend import Backends
from lib389.tests.topologies import topology_st

logging.getLogger(__name__).setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

DUMMY_BACKEND = 'dummydb'
DUMMY_BACKEND_DN = "cn=dummy_dn,cn=ldbm database,cn=plugins,cn=config"
NEW_SUFFIX_1_RDN = "o=new_suffix_1"
NEW_SUFFIX_2_RDN = "o=new_suffix_2"
BACKEND_NAME_1 = "new_backend_1"
BACKEND_NAME_2 = "new_backend_2"


@pytest.fixture()
def backend(topology_st, request):
    """Create and remove a backend"""

    log.info('Create a backend')
    backends = Backends(topology_st.standalone)
    backend = backends.create(properties={'nsslapd-suffix': NEW_SUFFIX_1_RDN,
                                          'cn': BACKEND_NAME_1})

    def fin():
        log.info("Just make it clean in the end")
        if backend.exists():
            backend.delete()
    request.addfinalizer(fin)

    return backend


def test_list(topology_st):
    """Test basic list method functionality

    :id: 084c0937-0b39-4e89-8561-081ae2b144c6
    :setup: Standalone instance
    :steps:
        1. List all backends
        2. Create a backend
        3. List all backends
        4. Create one more backend
        5. List all backends
        6. Clean up the created backends
    :expectedresults:
        1. Operation should be successful
        2. Backend should be created
        3. Created backend should be listed
        4. Backend should be created
        5. Created backend should be listed
        6. Operation should be successful
    """

    backends = Backends(topology_st.standalone)

    ents = backends.list()
    nb_backend = len(ents)
    for ent in ents:
        topology_st.standalone.log.info("List(%d): backend %s" %
                                        (nb_backend, ent.dn))

    log.info("Create a first backend and check list all backends")
    b1 = backends.create(properties={'cn': BACKEND_NAME_1,
                                     'nsslapd-suffix': NEW_SUFFIX_1_RDN})

    ents = backends.list()
    for ent in ents:
        topology_st.standalone.log.info("List(%d): backend %s" %
                                        (nb_backend + 1, ent.dn))
    assert len(ents) == (nb_backend + 1)

    log.info("Create a second backend and check list all backends")
    b2 = backends.create(properties={'cn': BACKEND_NAME_2,
                                     'nsslapd-suffix': NEW_SUFFIX_2_RDN})

    ents = backends.list()
    for ent in ents:
        topology_st.standalone.log.info("List(%d): backend %s" %
                                        (nb_backend + 2, ent.dn))
    assert len(ents) == (nb_backend + 2)

    log.info("Just make it clean in the end")
    b1.delete()
    b2.delete()


def test_create(topology_st):
    """Test basic list method functionality

    :id: df55a60b-f4dd-4f18-975d-4b223e63091f
    :setup: Standalone instance
    :steps:
        2. Create a backend specifying properties with a name and a suffix
        2. Create a backend specifying no properties
        3. Create a backend specifying suffix that already exist
        4. Create a backend specifying existing backend name but new suffix
        5. Create a backend specifying no backend name
        6. Create a backend specifying no backend suffix
        7. Clean up the created backend
    :expectedresults:
        1. Backend should be created
        2. Unwilling to perform error should be raised
        3. Unwilling to perform error should be raised
        4. Unwilling to perform error should be raised
        5. Unwilling to perform error should be raised
        6. Unwilling to perform error should be raised
        7. Operation should be successful
    """

    backends = Backends(topology_st.standalone)

    log.info("Create a backend")
    backend = backends.create(properties={'cn': BACKEND_NAME_1,
                                          'nsslapd-suffix': NEW_SUFFIX_1_RDN})

    log.info("Check behaviour with missing properties")
    with pytest.raises(ldap.UNWILLING_TO_PERFORM):
        backends.create()

    log.info("Check behaviour with already existing backend for that suffix")
    with pytest.raises(ldap.UNWILLING_TO_PERFORM):
        backends.create(properties={'cn': BACKEND_NAME_2,
                                    'nsslapd-suffix': NEW_SUFFIX_1_RDN})

    log.info("Check behaviour with already existing backend nasme, "
             "but new suffix")
    with pytest.raises(ldap.UNWILLING_TO_PERFORM):
        backends.create(properties={'cn': BACKEND_NAME_1,
                                    'nsslapd-suffix': NEW_SUFFIX_2_RDN})

    log.info("Create a backend without BACKEND_NAME")
    with pytest.raises(ldap.UNWILLING_TO_PERFORM):
        backends.create(properties={'nsslapd-suffix': NEW_SUFFIX_2_RDN})
        ents = backends.list()
        assert len(ents) == 1

    log.info("Create a backend without BACKEND_SUFFIX")
    with pytest.raises(ldap.UNWILLING_TO_PERFORM):
        backends.create(properties={'cn': BACKEND_NAME_1})
        ents = backends.list()
        assert len(ents) == 1

    log.info("Just make it clean in the end")
    backend.delete()


def test_get_valid(topology_st, backend):
    """Test basic get method functionality

    :id: d3e5ebf2-5598-41cd-b1fa-ac18d3057f80
    :setup: Standalone instance
    :steps:
        1. Get the backend with suffix
        2. Get the backend with backend name
        3. Get the backend with backend DN
    :expectedresults:
        1. Operation should be successful
        2. Operation should be successful
        3. Operation should be successful
    """

    backends = Backends(topology_st.standalone)

    log.info("Try to get a backend with backend suffix")
    d1 = backends.get(NEW_SUFFIX_1_RDN)
    assert d1 is not None
    topology_st.standalone.log.info("Suffix (%d): backend %s" % (1, d1.dn))

    log.info("Try to get a backend with backend name")
    d2 = backends.get(BACKEND_NAME_1)
    assert d2 is not None
    topology_st.standalone.log.info("Backend (%d): backend %s" % (1, d2.dn))

    log.info("Try to get a backend with backend DN")
    d3 = backends.get(dn=backend.dn)
    assert d3 is not None
    topology_st.standalone.log.info("DN (%d): backend %s" % (1, d3.dn))


def test_get_invalid(topology_st, backend):
    """Test the invalid situations while using get method

    :id: c5028350-1381-4d6d-82b8-4959a9b82964
    :setup: Standalone instance
    :steps:
        1. Get the backend with invalid suffix
        2. Get the backend with invalid backend name
        3. Get the backend with invalid backend DN
    :expectedresults:
        1. No such object error should be raised
        2. No such object error should be raised
        3. No such object error should be raised
    """

    backends = Backends(topology_st.standalone)

    log.info("Try to get the backend with invalid backend suffix")
    with pytest.raises(ldap.NO_SUCH_OBJECT):
        backends.get(NEW_SUFFIX_2_RDN)

    log.info("Try to get the backend with invalid backend name")
    with pytest.raises(ldap.NO_SUCH_OBJECT):
        backends.get(DUMMY_BACKEND)

    log.info("Try to get the backend with invalid backend DN")
    with pytest.raises(ldap.NO_SUCH_OBJECT):
        backends.get(dn=DUMMY_BACKEND_DN)


def test_delete(topology_st, backend):
    """Delete the backend and check that mapping tree and index were deleted too

    :id: d44dac3a-dae8-48e8-bd43-5be15237d093
    :setup: Standalone instance
    :steps:
        1. Create a backend
        2. Delete the backend
        3. Check all backend indexes were deleted
        4. Check backend mapping tree was deleted
    :expectedresults:
        1. Operation should be successful
        2. Operation should be successful
        3. There should be no backend indexes
        4. There should be no backend mapping tree
    """

    log.info('Delete a backend')
    backend.delete()

    log.info("Check that all indices are deleted")
    indexes = Indexes(topology_st.standalone, "cn=index,{}".format(backend.dn))
    assert not indexes.list()

    with pytest.raises(ldap.NO_SUCH_OBJECT):
        mts = MappingTrees(topology_st.standalone)
        mts.get(BACKEND_NAME_1)


def test_lint(topology_st, backend):
    """Test basic lint method functionality

    :id: 79e23980-e764-4839-b040-667357195710
    :setup: Standalone instance
    :steps:
        1. Create a backend
        2. Remove its mapping tree
        3. Run lint method on the backend
    :expectedresults:
        1. Operation should be successful
        2. Operation should be successful
        3. It should detect the missing mapping tree
    """

    log.info('Remove the mapping tree')
    mps = MappingTrees(topology_st.standalone)
    mps.delete(suffix=NEW_SUFFIX_1_RDN, bename=BACKEND_NAME_1)

    error_found = False
    for item in backend.lint():
        if item['dsle'] == 'DSBLE0001':
            error_found = True
    assert error_found


def test_create_sample_entries(topology_st, backend):
    """Test basic create_sample_entries method functionality

    :id: 9c995027-4884-494a-9af0-c8d2f43a0123
    :setup: Standalone instance
    :steps:
        1. Create a backend
        2. Create entries using create_sample_entries method
        3. Search the created entries (pick ou=People rdn)
    :expectedresults:
        1. Operation should be successful
        2. Operation should be successful
        3. The entry should be found
    """

    log.info('Create sample entries')
    backend.create_sample_entries(version='001003006')

    log.info('Search the entries')
    entries = topology_st.standalone.search_s('ou=people,{}'.format(NEW_SUFFIX_1_RDN), ldap.SCOPE_SUBTREE,
                                              '(objectclass=*)')
    assert entries


def test_get_monitor(topology_st, backend):
    """Test basic get_monitor method functionality

    :id: 7e22f967-d1a6-4e84-9659-f708fdd0b1ed
    :setup: Standalone instance
    :steps:
        1. Create a backend
        2. Get a MonitorBackend object instance using get_monitor function
        3. Directly define a MonitorBackend instance
        4. Assert that entries DN and attributes are the same
    :expectedresults:
        1. Operation should be successful
        2. Operation should be successful
        3. Operation should be successful
        4. The entries DN and attribute should be the same
    """

    log.info('Use get_monitor method to get MonitorBackend object')
    backend_monitor = backend.get_monitor()

    log.info('Directly define a MonitorBackend instance')
    monitor = MonitorBackend(topology_st.standalone, "cn=monitor,{}".format(backend.dn))

    log.info('Check the objects are the same')
    assert backend_monitor.dn == monitor.dn


def test_get_indexes(topology_st, backend):
    """Test basic get_indexes method functionality

    :id: 4e01d9e8-c355-4dd4-b7d9-a1d26afed768
    :setup: Standalone instance
    :steps:
        1. Create a backend
        2. Get an Indexes object instance using get_indexes function
        3. Directly define an Indexes instance
        4. Assert that the objects are the same
    :expectedresults:
        1. Operation should be successful
        2. Operation should be successful
        3. Operation should be successful
        4. The entries should match
    """

    log.info('Use get_indexes method to get Indexes object')
    backend_indexes = backend.get_indexes()

    log.info('Directly define an Indexes instance')
    indexes = Indexes(topology_st.standalone, "cn=index,{}".format(backend.dn))

    log.info('Check the objects are the same')
    index_found = False
    for i, j in zip(indexes.list(), backend_indexes.list()):
        if i.dn == j.dn:
            index_found = True
    assert index_found


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389 import DirSrv, Entry
import pytest
import os
import logging

from lib389.tests.topologies import topology_st as topo

from lib389.nss_ssl import NssSsl

DEBUGGING = os.getenv('DEBUGGING', False)

if DEBUGGING:
    logging.getLogger(__name__).setLevel(logging.DEBUG)
else:
    logging.getLogger(__name__).setLevel(logging.INFO)

log = logging.getLogger(__name__)


def test_external_ca():
    """Test the behaviour of our system ca database.

    :id: 321c85c1-9cf3-413f-9e26-99a8b327509c

    :steps:
        1. Create the system CA.
        2. Create an nss db
        3. Submit a CSR to the system ca
        4. Import the crt to the db
    :expectedresults:
        1. It works.
        2. It works.
        3. It works.
        4. It works.
    """
    # If it doesn't exist, create a cadb.
    ssca = NssSsl(dbpath='/tmp/lib389-ssca')
    ssca.reinit()
    ssca.create_rsa_ca()

    # Create certificate database.
    tlsdb = NssSsl(dbpath='/tmp/lib389-tlsdb')
    tlsdb.reinit()

    csr = tlsdb.create_rsa_key_and_csr()
    (ca, crt) = ssca.rsa_ca_sign_csr(csr)
    tlsdb.import_rsa_crt(ca, crt)


def test_nss_ssca_users(topo):
    """Validate that we can submit user certs to the ds ca for signing.

    :id: a47e47ed-2056-440b-8797-d13fa89098f6
    :steps:
        1. Find the ssca path.
        2. Assert it exists
        3. Create user certificates from the ssca
    :expectedresults:
        1. It works.
        2. It works.
        3. It works.
    """

    ssca = NssSsl(dbpath=topo.standalone.get_ssca_dir())

    if not ssca._db_exists():
        ssca.reinit()
        if not ssca._rsa_ca_exists():
            ssca.create_rsa_ca()

    # It better exist now!
    assert(ssca._rsa_ca_exists() is True)

    # Check making users certs. They should never conflict
    for user in ('william', 'noriko', 'mark'):
        # Create the user cert
        assert(ssca.create_rsa_user(user) is not None)
        # Assert it exists now
        assert(ssca._rsa_user_exists(user) is True)
    assert(ssca._rsa_user_exists('non_existen') is False)


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -vv %s" % CURRENT_FILE)

